@@ -28,7 +28,7 @@
 
 #include <compiz-plugin.h>
 
-#define CORE_ABIVERSION 20080214
+#define CORE_ABIVERSION 20080424
 
 #include <stdio.h>
 #include <sys/time.h>
@@ -346,18 +346,22 @@ compObjectFind (CompObject     *parent,
 
 /* session.c */
 
-typedef void (*SessionSaveYourselfProc) (CompCore   *c,
-					 const char *clientId,
-					 int        saveType,
-					 int        interactStyle,
-					 Bool       shutdown,
-					 Bool       fast);
+typedef enum {
+    CompSessionEventSaveYourself = 0,
+    CompSessionEventSaveComplete,
+    CompSessionEventDie,
+    CompSessionEventShutdownCancelled
+} CompSessionEvent;
 
-typedef void (*SessionDieProc) (CompCore *c);
+typedef enum {
+    CompSessionClientId = 0,
+    CompSessionPrevClientId
+} CompSessionClientIdType;
 
-typedef void (*SessionSaveCompleteProc) (CompCore *c);
-
-typedef void (*SessionShutdownCancelledProc) (CompCore *c);
+typedef void (*SessionEventProc) (CompCore         *c,
+				  CompSessionEvent event,
+				  CompOption       *arguments,
+				  unsigned int     nArguments);
 
 void
 initSession (char *smPrevClientId);
@@ -366,21 +370,13 @@ void
 closeSession (void);
 
 void
-sessionSaveYourself (CompCore   *c,
-		     const char *clientId,
-		     int        saveType,
-		     int        interactStyle,
-		     Bool       shutdown,
-		     Bool       fast);
+sessionEvent (CompCore         *c,
+	      CompSessionEvent event,
+	      CompOption       *arguments,
+	      unsigned int     nArguments);
 
-void
-sessionDie (CompCore *c);
-
-void
-sessionSaveComplete (CompCore *c);
-
-void
-sessionShutdownCancelled (CompCore *c);
+char *
+getSessionClientId (CompSessionClientIdType type);
 
 /* option.c */
 
@@ -403,7 +399,8 @@ typedef enum {
     CompActionStateTermEdgeDnd = 1 <<  8,
     CompActionStateCommit      = 1 <<  9,
     CompActionStateCancel      = 1 << 10,
-    CompActionStateAutoGrab    = 1 << 11
+    CompActionStateAutoGrab    = 1 << 11,
+    CompActionStateNoEdgeDelay = 1 << 12
 } CompActionState;
 
 typedef enum {
@@ -685,10 +682,7 @@ struct _CompCore {
     FileWatchAddedProc   fileWatchAdded;
     FileWatchRemovedProc fileWatchRemoved;
 
-    SessionSaveYourselfProc      sessionSaveYourself;
-    SessionDieProc               sessionDie;
-    SessionSaveCompleteProc      sessionSaveComplete;
-    SessionShutdownCancelledProc sessionShutdownCancelled;
+    SessionEventProc sessionEvent;
 };
 
 int
@@ -804,7 +798,8 @@ removeFileWatch (CompFileWatchHandle handle);
 #define COMP_DISPLAY_OPTION_TERMINAL			     64
 #define COMP_DISPLAY_OPTION_RUN_TERMINAL_KEY		     65
 #define COMP_DISPLAY_OPTION_PING_DELAY			     66
-#define COMP_DISPLAY_OPTION_NUM				     67
+#define COMP_DISPLAY_OPTION_EDGE_DELAY                       67
+#define COMP_DISPLAY_OPTION_NUM				     68
 
 typedef void (*HandleEventProc) (CompDisplay *display,
 				 XEvent	     *event);
@@ -1079,6 +1074,8 @@ struct _CompDisplay {
     CompTimeoutHandle autoRaiseHandle;
     Window	      autoRaiseWindow;
 
+    CompTimeoutHandle edgeDelayHandle;
+
     CompOptionValue plugin;
     Bool	    dirtyPluginList;
 
@@ -1260,6 +1257,17 @@ findCursorAtDisplay (CompDisplay *display);
 
 /* event.c */
 
+typedef struct _CompDelayedEdgeSettings
+{
+    CompDisplay *d;
+
+    unsigned int edge;
+    unsigned int state;
+
+    CompOption   option[7];
+    unsigned int nOption;
+} CompDelayedEdgeSettings;
+
 void
 handleEvent (CompDisplay *display,
 	     XEvent      *event);
@@ -1362,6 +1370,8 @@ typedef void (*DonePaintScreenProc) (CompScreen *screen);
 #define PAINT_SCREEN_TRANSFORMED_MASK		   (1 << 2)
 #define PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK (1 << 3)
 #define PAINT_SCREEN_CLEAR_MASK			   (1 << 4)
+#define PAINT_SCREEN_NO_OCCLUSION_DETECTION_MASK   (1 << 5)
+#define PAINT_SCREEN_NO_BACKGROUND_MASK            (1 << 6)
 
 typedef void (*PaintScreenProc) (CompScreen   *screen,
 				 CompOutput   *outputs,
@@ -1492,12 +1502,6 @@ typedef void (*PaintCursorProc) (CompCursor	     *cursor,
 				 Region		     region,
 				 unsigned int	     mask);
 
-#define PAINT_BACKGROUND_ON_TRANSFORMED_SCREEN_MASK (1 << 0)
-
-typedef void (*PaintBackgroundProc) (CompScreen   *screen,
-				     Region	  region,
-				     unsigned int mask);
-
 void
 preparePaintScreen (CompScreen *screen,
 		    int	       msSinceLastPaint);
@@ -1596,12 +1600,6 @@ paintCursor (CompCursor		 *cursor,
 	     Region		 region,
 	     unsigned int	 mask);
 
-void
-paintBackground (CompScreen   *screen,
-		 Region	      region,
-		 unsigned int mask);
-
-
 /* texture.c */
 
 #define POWER_OF_TWO(v) ((v & (v - 1)) == 0)
@@ -1713,10 +1711,14 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS	  9
 #define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  10
 #define COMP_SCREEN_OPTION_OUTPUTS		  11
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 12
-#define COMP_SCREEN_OPTION_OPACITY_MATCHES	  13
-#define COMP_SCREEN_OPTION_OPACITY_VALUES	  14
-#define COMP_SCREEN_OPTION_NUM		          15
+#define COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS	  12
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_LEVEL 13
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 14
+#define COMP_SCREEN_OPTION_OPACITY_MATCHES	  15
+#define COMP_SCREEN_OPTION_OPACITY_VALUES	  16
+#define COMP_SCREEN_OPTION_TEXTURE_COMPRESSION	  17
+#define COMP_SCREEN_OPTION_FORCE_INDEPENDENT      18
+#define COMP_SCREEN_OPTION_NUM		          19
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1739,6 +1741,17 @@ disableTexture (CompScreen  *screen,
 #define GLX_FRONT_LEFT_EXT                 0x20DE
 #endif
 
+#define OUTPUT_OVERLAP_MODE_SMART          0
+#define OUTPUT_OVERLAP_MODE_PREFER_LARGER  1
+#define OUTPUT_OVERLAP_MODE_PREFER_SMALLER 2
+#define OUTPUT_OVERLAP_MODE_LAST           OUTPUT_OVERLAP_MODE_PREFER_SMALLER
+
+#define FOCUS_PREVENTION_LEVEL_NONE     0
+#define FOCUS_PREVENTION_LEVEL_LOW      1
+#define FOCUS_PREVENTION_LEVEL_HIGH     2
+#define FOCUS_PREVENTION_LEVEL_VERYHIGH 3
+#define FOCUS_PREVENTION_LEVEL_LAST     FOCUS_PREVENTION_LEVEL_VERYHIGH
+
 typedef void (*FuncPtr) (void);
 typedef FuncPtr (*GLXGetProcAddressProc) (const GLubyte *procName);
 
@@ -1784,6 +1797,7 @@ typedef GLXPixmap (*GLXCreatePixmapProc) (Display     *display,
 
 typedef void (*GLActiveTextureProc) (GLenum texture);
 typedef void (*GLClientActiveTextureProc) (GLenum texture);
+typedef void (*GLMultiTexCoord2fProc) (GLenum, GLfloat, GLfloat);
 
 typedef void (*GLGenProgramsProc) (GLsizei n,
 				   GLuint  *programs);
@@ -1801,6 +1815,9 @@ typedef void (*GLProgramParameter4fProc) (GLenum  target,
 					  GLfloat y,
 					  GLfloat z,
 					  GLfloat w);
+typedef void (*GLGetProgramivProc) (GLenum target,
+				    GLenum pname,
+				    int    *params);
 
 typedef void (*GLGenFramebuffersProc) (GLsizei n,
 				       GLuint  *framebuffers);
@@ -1844,6 +1861,8 @@ typedef void (*GetAllowedActionsForWindowProc) (CompWindow   *w,
 
 typedef Bool (*FocusWindowProc) (CompWindow *window);
 
+typedef void (*ActivateWindowProc) (CompWindow *window);
+
 typedef Bool (*PlaceWindowProc) (CompWindow *window,
 				 int        x,
 				 int        y,
@@ -2032,6 +2051,7 @@ struct _CompScreen {
     int		      textureEnvCombine;
     int		      textureEnvCrossbar;
     int		      textureBorderClamp;
+    int		      textureCompression;
     GLint	      maxTextureSize;
     int		      fbo;
     int		      fragmentProgram;
@@ -2051,6 +2071,7 @@ struct _CompScreen {
     int	       nOutputDev;
     int	       currentOutputDev;
     CompOutput fullscreenOutput;
+    Bool       hasOverlappingOutputs;
 
     int windowOffsetX;
     int windowOffsetY;
@@ -2145,6 +2166,7 @@ struct _CompScreen {
 
     GLActiveTextureProc       activeTexture;
     GLClientActiveTextureProc clientActiveTexture;
+    GLMultiTexCoord2fProc     multiTexCoord2f;
 
     GLGenProgramsProc	     genPrograms;
     GLDeleteProgramsProc     deletePrograms;
@@ -2152,6 +2174,7 @@ struct _CompScreen {
     GLProgramStringProc	     programString;
     GLProgramParameter4fProc programEnvParameter4f;
     GLProgramParameter4fProc programLocalParameter4f;
+    GLGetProgramivProc       getProgramiv;
 
     GLGenFramebuffersProc        genFramebuffers;
     GLDeleteFramebuffersProc     deleteFramebuffers;
@@ -2172,7 +2195,6 @@ struct _CompScreen {
     EnableOutputClippingProc	    enableOutputClipping;
     DisableOutputClippingProc	    disableOutputClipping;
     ApplyScreenTransformProc	    applyScreenTransform;
-    PaintBackgroundProc		    paintBackground;
     PaintWindowProc		    paintWindow;
     DrawWindowProc		    drawWindow;
     AddWindowGeometryProc	    addWindowGeometry;
@@ -2181,6 +2203,7 @@ struct _CompScreen {
     GetOutputExtentsForWindowProc   getOutputExtentsForWindow;
     GetAllowedActionsForWindowProc  getAllowedActionsForWindow;
     FocusWindowProc		    focusWindow;
+    ActivateWindowProc              activateWindow;
     PlaceWindowProc                 placeWindow;
     ValidateWindowResizeRequestProc validateWindowResizeRequest;
 
@@ -2267,6 +2290,9 @@ showOutputWindow (CompScreen *s);
 void
 hideOutputWindow (CompScreen *s);
 
+void
+updateOutputWindow (CompScreen *s);
+
 Bool
 addScreen (CompDisplay *display,
 	   int	       screenNum,
@@ -2575,6 +2601,7 @@ struct _CompWindow {
     Bool	      redirected;
     Bool	      managed;
     Bool	      bindFailed;
+    Bool	      overlayWindow;
     int		      destroyRefCnt;
     int		      unmapRefCnt;
 
@@ -2650,6 +2677,7 @@ struct _CompWindow {
 
     GLfloat  *vertices;
     int      vertexSize;
+    int      vertexStride;
     GLushort *indices;
     int      indexSize;
     int      vCount;
@@ -2852,6 +2880,12 @@ configureXWindow (CompWindow	 *w,
 		  unsigned int	 valueMask,
 		  XWindowChanges *xwc);
 
+unsigned int
+adjustConfigureRequestForGravity (CompWindow     *w,
+				  XWindowChanges *xwc,
+				  unsigned int   xwcm,
+				  int            gravity);
+
 void
 moveResizeWindow (CompWindow     *w,
 		  XWindowChanges *xwc,
