@@ -1,7 +1,7 @@
 /****************************************************************************
 ** KWD::Window meta object code from reading C++ file 'window.h'
 **
-** Created: Tue Feb 5 17:07:16 2008
+** Created: Thu May 29 14:07:49 2008
 **      by: The Qt MOC ($Id: qt/moc_yacc.cpp   3.3.8   edited Feb 2 14:59 $)
 **
 ** WARNING! All changes made in this file will be lost!
@@ -14,7 +14,7 @@
 
 #include <private/qucomextra_p.h>
 #if !defined(Q_MOC_OUTPUT_REVISION) || (Q_MOC_OUTPUT_REVISION != 26)
-#error "This file was generated using the moc from 3.3.8. It"
+#error "This file was generated using the moc from 3.3.8b. It"
 #error "cannot be used with the include files from this version of Qt."
 #error "(The moc has changed too much.)"
 #endif
