@@ -49,6 +49,8 @@ typedef struct _BlurFunction {
     int target;
     int param;
     int unit;
+    int startTC;
+    int numITC;
 } BlurFunction;
 
 typedef struct _BlurBox {
@@ -101,7 +103,8 @@ typedef struct _BlurDisplay {
 #define BLUR_SCREEN_OPTION_MIPMAP_LOD        8
 #define BLUR_SCREEN_OPTION_SATURATION        9
 #define BLUR_SCREEN_OPTION_BLUR_OCCLUSION    10
-#define BLUR_SCREEN_OPTION_NUM		     11
+#define BLUR_SCREEN_OPTION_INDEPENDENT_TEX   11
+#define BLUR_SCREEN_OPTION_NUM		     12
 
 typedef struct _BlurScreen {
     int	windowPrivateIndex;
@@ -152,12 +155,15 @@ typedef struct _BlurScreen {
     int    height;
 
     GLuint program;
+    int    maxTemp;
     GLuint fbo;
     Bool   fboStatus;
 
     float amp[BLUR_GAUSSIAN_RADIUS_MAX];
     float pos[BLUR_GAUSSIAN_RADIUS_MAX];
     int	  numTexop;
+
+    CompTransform mvp;
 } BlurScreen;
 
 typedef struct _BlurWindow {
@@ -671,6 +677,17 @@ blurSetScreenOption (CompPlugin      *plugin,
 	    damageScreen (screen);
 	    return TRUE;
 	}
+    case BLUR_SCREEN_OPTION_INDEPENDENT_TEX:
+	if (compSetBoolOption (o, value))
+	{
+	    filter = bs->opt[BLUR_SCREEN_OPTION_FILTER].value.i;
+	    if (filter == BLUR_FILTER_GAUSSIAN)
+	    {
+		blurReset (screen);
+		damageScreen (screen);
+	    }
+	    return TRUE;
+	}
     default:
 	break;
     }
@@ -1075,7 +1092,9 @@ static int
 getDstBlurFragmentFunction (CompScreen  *s,
 			    CompTexture *texture,
 			    int	        param,
-			    int		unit)
+			    int		unit,
+			    int         numITC,
+			    int         startTC)
 {
     BlurFunction     *function;
     CompFunctionData *data;
@@ -1096,19 +1115,24 @@ getDstBlurFragmentFunction (CompScreen  *s,
     }
 
     for (function = bs->dstBlurFunctions; function; function = function->next)
-	if (function->param  == param  &&
-	    function->target == target &&
-	    function->unit   == unit)
+	if (function->param   == param  &&
+	    function->target  == target &&
+	    function->unit    == unit   &&
+	    function->numITC  == numITC &&
+	    function->startTC == startTC)
 	    return function->handle;
 
     data = createFunctionData ();
     if (data)
     {
-	static char *temp[] = { "coord", "mask", "sum", "dst" };
-	int	    i, handle = 0;
+	static char *temp[] = { "fCoord", "mask", "sum", "dst" };
+	int	    i, j, handle = 0;
 	char	    str[1024];
 	int	    saturation = bs->opt[BLUR_SCREEN_OPTION_SATURATION].value.i;
 	Bool	    ok = TRUE;
+	int	    numIndirect;
+	int	    numIndirectOp;
+	int	    base, end, ITCbase;
 
 	for (i = 0; i < sizeof (temp) / sizeof (temp[0]); i++)
 	    ok &= addTempHeaderOpToFunctionData (data, temp[i]);
@@ -1130,22 +1154,22 @@ getDstBlurFragmentFunction (CompScreen  *s,
 	    ok &= addColorOpToFunctionData (data, "output", "output");
 
 	    snprintf (str, 1024,
-		      "MUL coord, fragment.position, program.env[%d];",
+		      "MUL fCoord, fragment.position, program.env[%d];",
 		      param);
 
 	    ok &= addDataOpToFunctionData (data, str);
 
 	    snprintf (str, 1024,
-		      "ADD t0, coord, program.env[%d];"
+		      "ADD t0, fCoord, program.env[%d];"
 		      "TEX s0, t0, texture[%d], %s;"
 
-		      "SUB t1, coord, program.env[%d];"
+		      "SUB t1, fCoord, program.env[%d];"
 		      "TEX s1, t1, texture[%d], %s;"
 
-		      "MAD t2, program.env[%d], { -1.0, 1.0, 0.0, 0.0 }, coord;"
+		      "MAD t2, program.env[%d], { -1.0, 1.0, 0.0, 0.0 }, fCoord;"
 		      "TEX s2, t2, texture[%d], %s;"
 
-		      "MAD t3, program.env[%d], { 1.0, -1.0, 0.0, 0.0 }, coord;"
+		      "MAD t3, program.env[%d], { 1.0, -1.0, 0.0, 0.0 }, fCoord;"
 		      "TEX s3, t3, texture[%d], %s;"
 
 		      "MUL_SAT mask, output.a, program.env[%d];"
@@ -1164,25 +1188,51 @@ getDstBlurFragmentFunction (CompScreen  *s,
 	    ok &= addDataOpToFunctionData (data, str);
 	} break;
 	case BLUR_FILTER_GAUSSIAN: {
-	    static char *filterTemp[] = {
-		"tCoord", "pix"
-	    };
 
-	    for (i = 0; i < sizeof (filterTemp) / sizeof (filterTemp[0]); i++)
-		ok &= addTempHeaderOpToFunctionData (data, filterTemp[i]);
+	    /* try to use only half of the available temporaries to keep
+	       other plugins working */
+	    if ((bs->maxTemp / 2) - 4 >
+		 (bs->numTexop + (bs->numTexop - numITC)) * 2)
+	    {
+		numIndirect   = 1;
+		numIndirectOp = bs->numTexop;
+	    }
+	    else
+	    {
+		i = MAX(((bs->maxTemp / 2) - 4) / 4, 1);
+		numIndirect = ceil ((float)bs->numTexop / (float)i);
+		numIndirectOp = ceil ((float)bs->numTexop / (float)numIndirect);
+	    }
 
+	    /* we need to define all coordinate temporaries if we have
+	       multiple indirection steps */
+	    j = (numIndirect > 1) ? 0 : numITC;
+
+	    for (i = 0; i < numIndirectOp * 2; i++)
+	    {
+		snprintf (str, 1024, "pix_%d", i);
+		ok &= addTempHeaderOpToFunctionData (data, str);
+	    }
+
+	    for (i = j * 2; i < numIndirectOp * 2; i++)
+	    {
+		snprintf (str, 1024, "coord_%d", i);
+		ok &= addTempHeaderOpToFunctionData (data, str);
+	    }
+
+	    
 	    ok &= addFetchOpToFunctionData (data, "output", NULL, target);
 	    ok &= addColorOpToFunctionData (data, "output", "output");
 
 	    snprintf (str, 1024,
-		      "MUL coord, fragment.position, program.env[%d];",
+		      "MUL fCoord, fragment.position, program.env[%d];",
 		      param);
 
 	    ok &= addDataOpToFunctionData (data, str);
 
 	    snprintf (str, 1024,
-		      "TEX sum, coord, texture[%d], %s;",
-		      unit + 1, targetString);
+		      "TEX sum, fCoord, texture[%d], %s;",
+		       unit + 1, targetString);
 
 	    ok &= addDataOpToFunctionData (data, str);
 
@@ -1193,33 +1243,69 @@ getDstBlurFragmentFunction (CompScreen  *s,
 
 	    ok &= addDataOpToFunctionData (data, str);
 
-	    for (i = 0; i < bs->numTexop; i++)
+	    for (j = 0; j < numIndirect; j++)
 	    {
-		snprintf (str, 1024,
-			  "ADD tCoord, coord, program.env[%d];"
-			  "TEX pix, tCoord, texture[%d], %s;"
-			  "MAD sum, pix, %f, sum;"
-			  "SUB tCoord, coord, program.env[%d];"
-			  "TEX pix, tCoord, texture[%d], %s;"
-			  "MAD sum, pix, %f, sum;",
-			  param + 2 + i,
-			  unit + 1, targetString,
-			  bs->amp[i],
-			  param + 2 + i,
-			  unit + 1, targetString,
-			  bs->amp[i]);
+		base = j * numIndirectOp;
+		end  = MIN ((j + 1) * numIndirectOp, bs->numTexop) - base;
+		
+		ITCbase = MAX (numITC - base, 0);
 
-		ok &= addDataOpToFunctionData (data, str);
+		for (i = ITCbase; i < end; i++)
+		{
+		    snprintf (str, 1024,
+			      "ADD coord_%d, fCoord, {0.0, %g, 0.0, 0.0};"
+			      "SUB coord_%d, fCoord, {0.0, %g, 0.0, 0.0};",
+			      i * 2, bs->pos[base + i] * bs->ty,
+			      (i * 2) + 1, bs->pos[base + i] * bs->ty);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = 0; i < ITCbase; i++)
+		{
+		    snprintf (str, 1024,
+			"TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;"
+			"TXP pix_%d, fragment.texcoord[%d], texture[%d], %s;",
+			i * 2, startTC + ((i + base) * 2),
+			unit + 1, targetString,
+			(i * 2) + 1, startTC + 1 + ((i + base) * 2),
+			unit + 1, targetString);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = ITCbase; i < end; i++)
+		{
+		    snprintf (str, 1024,
+			      "TEX pix_%d, coord_%d, texture[%d], %s;"
+			      "TEX pix_%d, coord_%d, texture[%d], %s;",
+			      i * 2, i * 2,
+			      unit + 1, targetString,
+			      (i * 2) + 1, (i * 2) + 1,
+			      unit + 1, targetString);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
+
+		for (i = 0; i < end * 2; i++)
+		{
+		    snprintf (str, 1024,
+			      "MAD sum, pix_%d, %f, sum;",
+			      i, bs->amp[base + (i / 2)]);
+
+		    ok &= addDataOpToFunctionData (data, str);
+		}
 	    }
+
 	} break;
 	case BLUR_FILTER_MIPMAP:
 	    ok &= addFetchOpToFunctionData (data, "output", NULL, target);
 	    ok &= addColorOpToFunctionData (data, "output", "output");
 
 	    snprintf (str, 1024,
-		      "MUL coord, fragment.position, program.env[%d].xyzz;"
-		      "MOV coord.w, program.env[%d].w;"
-		      "TXB sum, coord, texture[%d], %s;"
+		      "MUL fCoord, fragment.position, program.env[%d].xyzz;"
+		      "MOV fCoord.w, program.env[%d].w;"
+		      "TXB sum, fCoord, texture[%d], %s;"
 		      "MUL_SAT mask, output.a, program.env[%d];",
 		      param, param, unit, targetString,
 		      param + 1);
@@ -1258,10 +1344,12 @@ getDstBlurFragmentFunction (CompScreen  *s,
 	{
 	    handle = createFragmentFunction (s, "blur", data);
 
-	    function->handle = handle;
-	    function->target = target;
-	    function->param  = param;
-	    function->unit   = unit;
+	    function->handle  = handle;
+	    function->target  = target;
+	    function->param   = param;
+	    function->unit    = unit;
+	    function->numITC  = numITC;
+	    function->startTC = startTC;
 
 	    function->next = bs->dstBlurFunctions;
 	    bs->dstBlurFunctions = function;
@@ -1302,7 +1390,7 @@ projectVertices (CompScreen	     *s,
 
     while (n--)
     {
-	if (!gluProject (object[0], object[1], 0.0,
+	if (!gluProject (object[0], object[1], object[2],
 			 dModel, dProjection, viewport,
 			 &x, &y, &z))
 	    return FALSE;
@@ -1310,7 +1398,7 @@ projectVertices (CompScreen	     *s,
 	screen[0] = x;
 	screen[1] = y;
 
-	object += 2;
+	object += 3;
 	screen += 2;
     }
 
@@ -1339,7 +1427,7 @@ loadFragmentProgram (CompScreen *s,
     if (glGetError () != GL_NO_ERROR || errorPos != -1)
     {
 	compLogMessage (s->display, "blur", CompLogLevelError,
-			"Failed to load blur program");
+			"Failed to load blur program %s", string);
 
 	(*s->deletePrograms) (1, program);
 	*program = 0;
@@ -1351,12 +1439,15 @@ loadFragmentProgram (CompScreen *s,
 }
 
 static Bool
-loadFilterProgram (CompScreen *s)
+loadFilterProgram (CompScreen *s, int numITC)
 {
-    char  buffer[2048];
+    char  buffer[4096];
     char  *targetString;
     char  *str = buffer;
-    int   i;
+    int   i, j;
+    int   numIndirect;
+    int   numIndirectOp;
+    int   base, end, ITCbase;
 
     BLUR_SCREEN (s);
 
@@ -1370,7 +1461,27 @@ loadFilterProgram (CompScreen *s)
 		    "ATTRIB texcoord = fragment.texcoord[0];"
 		    "TEMP sum;");
 
-    str += sprintf (str, "TEMP tCoord, pix;");
+    if (bs->maxTemp - 1 > (bs->numTexop + (bs->numTexop - numITC)) * 2)
+    {
+	numIndirect   = 1;
+	numIndirectOp = bs->numTexop;
+    }
+    else
+    {
+	i = (bs->maxTemp - 1) / 4;
+	numIndirect = ceil ((float)bs->numTexop / (float)i);
+	numIndirectOp = ceil ((float)bs->numTexop / (float)numIndirect);
+    }
+
+    /* we need to define all coordinate temporaries if we have
+       multiple indirection steps */
+    j = (numIndirect > 1) ? 0 : numITC;
+
+    for (i = 0; i < numIndirectOp; i++)
+	str += sprintf (str,"TEMP pix_%d, pix_%d;", i * 2, (i * 2) + 1);
+
+    for (i = j; i < numIndirectOp; i++)
+	str += sprintf (str,"TEMP coord_%d, coord_%d;", i * 2, (i * 2) + 1);
 
     str += sprintf (str,
 		    "TEX sum, texcoord, texture[0], %s;",
@@ -1380,16 +1491,39 @@ loadFilterProgram (CompScreen *s)
 		    "MUL sum, sum, %f;",
 		    bs->amp[bs->numTexop]);
 
-    for (i = 0; i < bs->numTexop; i++)
-	str += sprintf (str,
-			"ADD tCoord, texcoord, program.local[%d];"
-			"TEX pix, tCoord, texture[0], %s;"
-			"MAD sum, pix, %f, sum;"
-			"SUB tCoord, texcoord, program.local[%d];"
-			"TEX pix, tCoord, texture[0], %s;"
-			"MAD sum, pix, %f, sum;",
-			i, targetString, bs->amp[i],
-			i, targetString, bs->amp[i]);
+    for (j = 0; j < numIndirect; j++)
+    {
+	base = j * numIndirectOp;
+	end  = MIN ((j + 1) * numIndirectOp, bs->numTexop) - base;
+	
+	ITCbase = MAX (numITC - base, 0);
+
+	for (i = ITCbase; i < end; i++)
+	    str += sprintf (str,
+			    "ADD coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};"
+			    "SUB coord_%d, texcoord, {%g, 0.0, 0.0, 0.0};",
+			    i * 2, bs->pos[base + i] * bs->tx,
+			    (i * 2) + 1, bs->pos[base + i] * bs->tx);
+
+	for (i = 0; i < ITCbase; i++)
+	    str += sprintf (str,
+		"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;"
+		"TEX pix_%d, fragment.texcoord[%d], texture[0], %s;",
+		i * 2, ((i + base) * 2) + 1, targetString,
+		(i * 2) + 1, ((i + base) * 2) + 2, targetString);
+
+	for (i = ITCbase; i < end; i++)
+	    str += sprintf (str,
+			    "TEX pix_%d, coord_%d, texture[0], %s;"
+			    "TEX pix_%d, coord_%d, texture[0], %s;",
+			    i * 2, i * 2, targetString,
+			    (i * 2) + 1, (i * 2) + 1, targetString);
+
+	for (i = 0; i < end * 2; i++)
+	    str += sprintf (str,
+			    "MAD sum, pix_%d, %f, sum;",
+			    i, bs->amp[base + (i / 2)]);
+    }
 
     str += sprintf (str,
 		    "MOV result.color, sum;"
@@ -1484,17 +1618,24 @@ fboUpdate (CompScreen *s,
 	   BoxPtr     pBox,
 	   int	      nBox)
 {
-    int i, y;
+    int  i, y, iTC = 0;
+    Bool wasCulled = glIsEnabled (GL_CULL_FACE);
 
     BLUR_SCREEN (s);
 
+    if (s->maxTextureUnits && 
+	bs->opt[BLUR_SCREEN_OPTION_INDEPENDENT_TEX].value.b)
+	iTC = MIN ((s->maxTextureUnits - 1) / 2, bs->numTexop);
+
     if (!bs->program)
-	if (!loadFilterProgram (s))
+	if (!loadFilterProgram (s, iTC))
 	    return FALSE;
 
     if (!fboPrologue (s))
 	return FALSE;
 
+    glDisable (GL_CULL_FACE);
+
     glDisableClientState (GL_TEXTURE_COORD_ARRAY);
 
     glBindTexture (bs->target, bs->texture[0]);
@@ -1502,26 +1643,63 @@ fboUpdate (CompScreen *s,
     glEnable (GL_FRAGMENT_PROGRAM_ARB);
     (*s->bindProgram) (GL_FRAGMENT_PROGRAM_ARB, bs->program);
 
-    for (i = 0; i < bs->numTexop; i++)
-	(*s->programLocalParameter4f) (GL_FRAGMENT_PROGRAM_ARB, i,
-				       bs->tx * bs->pos[i],
-				       0.0f, 0.0f, 0.0f);
-
     glBegin (GL_QUADS);
 
     while (nBox--)
     {
 	y = s->height - pBox->y2;
 
+	for (i = 0; i < iTC; i++)
+	{
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2),
+				   bs->tx * (pBox->x1 + bs->pos[i]),
+				   bs->ty * y);
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2) + 1,
+				   bs->tx * (pBox->x1 - bs->pos[i]),
+				   bs->ty * y);
+	}
+
 	glTexCoord2f (bs->tx * pBox->x1, bs->ty * y);
 	glVertex2i   (pBox->x1, y);
+
+	for (i = 0; i < iTC; i++)
+	{
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2),
+				   bs->tx * (pBox->x2 + bs->pos[i]),
+				   bs->ty * y);
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2) + 1,
+				   bs->tx * (pBox->x2 - bs->pos[i]),
+				   bs->ty * y);
+	}
+
 	glTexCoord2f (bs->tx * pBox->x2, bs->ty * y);
 	glVertex2i   (pBox->x2, y);
 
 	y = s->height - pBox->y1;
 
+	for (i = 0; i < iTC; i++)
+	{
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2),
+				   bs->tx * (pBox->x2 + bs->pos[i]),
+				   bs->ty * y);
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2) + 1,
+				   bs->tx * (pBox->x2 - bs->pos[i]),
+				   bs->ty * y);
+	}
+
 	glTexCoord2f (bs->tx * pBox->x2, bs->ty * y);
 	glVertex2i   (pBox->x2, y);
+
+	for (i = 0; i < iTC; i++)
+	{
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2),
+				   bs->tx * (pBox->x1 + bs->pos[i]),
+				   bs->ty * y);
+	    (*s->multiTexCoord2f) (GL_TEXTURE1_ARB + (i * 2) + 1,
+				   bs->tx * (pBox->x1 - bs->pos[i]),
+				   bs->ty * y);
+	}
+
 	glTexCoord2f (bs->tx * pBox->x1, bs->ty * y);
 	glVertex2i   (pBox->x1, y);
 
@@ -1534,6 +1712,9 @@ fboUpdate (CompScreen *s,
 
     glEnableClientState (GL_TEXTURE_COORD_ARRAY);
 
+    if (wasCulled)
+	glEnable (GL_CULL_FACE);
+
     fboEpilogue (s);
 
     return TRUE;
@@ -1548,11 +1729,11 @@ blurProjectRegion (CompWindow	       *w,
 {
     CompScreen *s = w->screen;
     float      screen[MAX_VERTEX_PROJECT_COUNT * 2];
-    float      vertices[MAX_VERTEX_PROJECT_COUNT * 2];
-    int	       nVertices;
+    float      vertices[MAX_VERTEX_PROJECT_COUNT * 3];
+    int	       nVertices, nQuadCombine;
     int        i, j, stride;
     float      *v, *vert;
-    float      minX, maxX, minY, maxY;
+    float      minX, maxX, minY, maxY, minZ, maxZ;
     float      *scr;
     REGION     region;
 
@@ -1565,10 +1746,11 @@ blurProjectRegion (CompWindow	       *w,
     if (!w->vCount)
 	return;
 
-    nVertices = (w->indexCount) ? w->indexCount: w->vCount;
+    nVertices    = (w->indexCount) ? w->indexCount: w->vCount;
+    nQuadCombine = 1;
 
-    stride = (1 + w->texUnits) * 2;
-    vert = w->vertices + (stride - 2);
+    stride = w->vertexStride;
+    vert = w->vertices + (stride - 3);
 
     /* we need to find the best value here */
     if (nVertices <= MAX_VERTEX_PROJECT_COUNT)
@@ -1584,8 +1766,9 @@ blurProjectRegion (CompWindow	       *w,
 		v = vert + (stride * i);
 	    }
 
-	    vertices[i * 2] = v[0];
-	    vertices[(i * 2) + 1] = v[1];
+	    vertices[i * 3] = v[0];
+	    vertices[(i * 3) + 1] = v[1];
+	    vertices[(i * 3) + 2] = v[2];
 	}
     }
     else
@@ -1594,6 +1777,8 @@ blurProjectRegion (CompWindow	       *w,
 	maxX = 0;
 	minY = s->height;
 	maxY = 0;
+	minZ = 1000000;
+	maxZ = -1000000;
 
 	for (i = 0; i < w->vCount; i++)
 	{
@@ -1610,18 +1795,38 @@ blurProjectRegion (CompWindow	       *w,
 
 	    if (v[1] > maxY)
 		maxY = v[1];
+
+	    if (v[2] < minZ)
+		minZ = v[2];
+
+	    if (v[2] > maxZ)
+		maxZ = v[2];
 	}
 
-	vertices[0] = vertices[6] = minX;
-	vertices[1] = vertices[3] = minY;
-	vertices[2] = vertices[4] = maxX;
-	vertices[5] = vertices[7] = maxY;
+	vertices[0] = vertices[9]  = minX;
+	vertices[1] = vertices[4]  = minY;
+	vertices[3] = vertices[6]  = maxX;
+	vertices[7] = vertices[10] = maxY;
+	vertices[2] = vertices[5]  = maxZ;
+	vertices[8] = vertices[11] = maxZ;
 
 	nVertices = 4;
+
+	if (maxZ != minZ)
+	{
+	    vertices[12] = vertices[21] = minX;
+	    vertices[13] = vertices[16] = minY;
+	    vertices[15] = vertices[18] = maxX;
+	    vertices[19] = vertices[22] = maxY;
+	    vertices[14] = vertices[17] = minZ;
+	    vertices[20] = vertices[23] = minZ;
+	    nQuadCombine = 2;
+	}
     }
 
+
     if (!projectVertices (w->screen, output, transform, vertices, screen,
-			  nVertices))
+			  nVertices * nQuadCombine))
 	return;
 
     region.rects    = &region.extents;
@@ -1636,7 +1841,7 @@ blurProjectRegion (CompWindow	       *w,
 	minY = s->height;
 	maxY = 0;
 
-	for (j = 0; j < 8; j += 2)
+	for (j = 0; j < 8 * nQuadCombine; j += 2)
 	{
 	    if (scr[j] < minX)
 		minX = scr[j];
@@ -1911,6 +2116,12 @@ blurDrawWindow (CompWindow	     *w,
 	    Bool   clipped = FALSE;
 	    BoxRec box = { 0, 0, 0, 0 };
 	    Region reg;
+	    int    i;
+
+	    for (i = 0; i < 16; i++)
+		bs->mvp.m[i] = s->projection[i];
+
+	    matrixMultiply (&bs->mvp, &bs->mvp, transform);
 
 	    if (mask & PAINT_WINDOW_TRANSFORMED_MASK)
 		reg = &infiniteRegion;
@@ -2048,6 +2259,7 @@ blurDrawWindowTexture (CompWindow	    *w,
 	int	       param, function;
 	int	       unit = 0;
 	GLfloat	       dx, dy;
+	int            iTC = 0;
 
 	if (bw->blur)
 	{
@@ -2082,7 +2294,8 @@ blurDrawWindowTexture (CompWindow	    *w,
 		param = allocFragmentParameters (&dstFa, 3);
 		unit  = allocFragmentTextureUnits (&dstFa, 1);
 
-		function = getDstBlurFragmentFunction (s, texture, param, unit);
+		function = 
+		    getDstBlurFragmentFunction (s, texture, param, unit, 0, 0);
 		if (function)
 		{
 		    addFragmentFunction (&dstFa, function);
@@ -2105,13 +2318,23 @@ blurDrawWindowTexture (CompWindow	    *w,
 		}
 		break;
 	    case BLUR_FILTER_GAUSSIAN:
-		param = allocFragmentParameters (&dstFa, 5);
+		if (bs->opt[BLUR_SCREEN_OPTION_INDEPENDENT_TEX].value.b)
+		{
+		    /* leave one free texture unit for fragment position */
+		    iTC = MAX (0, s->maxTextureUnits - (w->texUnits + 1));
+		    if (iTC)
+			iTC = MIN (iTC / 2, bs->numTexop);
+		}
+
+		param = allocFragmentParameters (&dstFa, 2);
 		unit  = allocFragmentTextureUnits (&dstFa, 2);
 
-		function = getDstBlurFragmentFunction (s, texture, param, unit);
+		function = 
+		    getDstBlurFragmentFunction (s, texture, param, unit, 
+						iTC, w->texUnits);
 		if (function)
 		{
-		    int i;
+		    int           i;
 
 		    addFragmentFunction (&dstFa, function);
 
@@ -2131,19 +2354,108 @@ blurDrawWindowTexture (CompWindow	    *w,
 						 threshold, threshold,
 						 threshold, threshold);
 
+		    if (iTC)
+		    {
+			CompTransform tm, rm;
+		        float s_gen[4], t_gen[4], q_gen[4];
+
+			for (i = 0; i < 16; i++)
+			    tm.m[i] = 0;
+			tm.m[0] = (bs->output->width / 2.0) * bs->tx;
+			tm.m[5] = (bs->output->height / 2.0) * bs->ty;
+			tm.m[10] = 1;
+
+			tm.m[12] = (bs->output->width / 2.0 +
+			bs->output->region.extents.x1) * bs->tx;
+			tm.m[13] = (bs->output->height / 2.0 + s->height -
+				bs->output->region.extents.y2) * bs->ty;
+			tm.m[14] = 1;
+			tm.m[15] = 1;
+
+			matrixMultiply (&tm, &tm, &bs->mvp);
+
+			for (i = 0; i < iTC; i++)
+			{
+			    (*s->activeTexture) (GL_TEXTURE0_ARB + w->texUnits
+						 + (i * 2));
+
+			    matrixGetIdentity (&rm);
+			    rm.m[13] = bs->ty * bs->pos[i];
+			    matrixMultiply (&rm, &rm, &tm);
+			    s_gen[0] = rm.m[0];
+			    s_gen[1] = rm.m[4];
+			    s_gen[2] = rm.m[8];
+			    s_gen[3] = rm.m[12];
+			    t_gen[0] = rm.m[1];
+			    t_gen[1] = rm.m[5];
+			    t_gen[2] = rm.m[9];
+			    t_gen[3] = rm.m[13];
+			    q_gen[0] = rm.m[3];
+			    q_gen[1] = rm.m[7];
+			    q_gen[2] = rm.m[11];
+			    q_gen[3] = rm.m[15];
+
+			    glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen);
+			    glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen);
+			    glTexGenfv(GL_Q, GL_OBJECT_PLANE, q_gen);
+
+			    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+			    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+			    glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+
+			    glEnable(GL_TEXTURE_GEN_S);
+			    glEnable(GL_TEXTURE_GEN_T);
+			    glEnable(GL_TEXTURE_GEN_Q);
+
+			    (*s->activeTexture) (GL_TEXTURE0_ARB + w->texUnits
+						 + 1 + (i * 2));
+
+			    matrixGetIdentity (&rm);
+			    rm.m[13] = -bs->ty * bs->pos[i];
+			    matrixMultiply (&rm, &rm, &tm);
+			    s_gen[0] = rm.m[0];
+			    s_gen[1] = rm.m[4];
+			    s_gen[2] = rm.m[8];
+			    s_gen[3] = rm.m[12];
+			    t_gen[0] = rm.m[1];
+			    t_gen[1] = rm.m[5];
+			    t_gen[2] = rm.m[9];
+			    t_gen[3] = rm.m[13];
+			    q_gen[0] = rm.m[3];
+			    q_gen[1] = rm.m[7];
+			    q_gen[2] = rm.m[11];
+			    q_gen[3] = rm.m[15];
+
+			    glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen);
+			    glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen);
+			    glTexGenfv(GL_Q, GL_OBJECT_PLANE, q_gen);
+
+			    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+			    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+			    glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE,
+				      GL_OBJECT_LINEAR);
+
+			    glEnable(GL_TEXTURE_GEN_S);
+			    glEnable(GL_TEXTURE_GEN_T);
+			    glEnable(GL_TEXTURE_GEN_Q);
+			}
+			
+			(*s->activeTexture) (GL_TEXTURE0_ARB);
+		    }
 
-		    for (i = 0; i < bs->numTexop; i++)
-			(*s->programEnvParameter4f) (GL_FRAGMENT_PROGRAM_ARB,
-						     param + 2 + i,
-						     0.0f, bs->ty * bs->pos[i],
-						     0.0f, 0.0f);
 		}
 		break;
 	    case BLUR_FILTER_MIPMAP:
 		param = allocFragmentParameters (&dstFa, 2);
 		unit  = allocFragmentTextureUnits (&dstFa, 1);
 
-		function = getDstBlurFragmentFunction (s, texture, param, unit);
+		function = 
+		    getDstBlurFragmentFunction (s, texture, param, unit, 0, 0);
 		if (function)
 		{
 		    float lod = bs->opt[BLUR_SCREEN_OPTION_MIPMAP_LOD].value.f;
@@ -2210,6 +2522,19 @@ blurDrawWindowTexture (CompWindow	    *w,
 	    glBindTexture (bs->target, 0);
 	    (*s->activeTexture) (GL_TEXTURE0_ARB);
 	}
+	
+	if (iTC)
+	{
+	    int i;
+	    for (i = w->texUnits; i < w->texUnits + (2 * iTC); i++)
+	    {
+		(*s->activeTexture) (GL_TEXTURE0_ARB + i);
+		glDisable(GL_TEXTURE_GEN_S);
+		glDisable(GL_TEXTURE_GEN_T);
+		glDisable(GL_TEXTURE_GEN_Q);
+	    }
+	    (*s->activeTexture) (GL_TEXTURE0_ARB);
+	}
     }
     else
     {
@@ -2555,7 +2880,8 @@ static const CompMetadataOptionInfo blurScreenOptionInfo[] = {
     { "gaussian_strength", "float", "<min>0.0</min><max>1.0</max>", 0, 0 },
     { "mipmap_lod", "float", "<min>0.1</min><max>5.0</max>", 0, 0 },
     { "saturation", "int", "<min>0</min><max>100</max>", 0, 0 },
-    { "occlusion", "bool", 0, 0, 0 }
+    { "occlusion", "bool", 0, 0, 0 },
+    { "independent_tex", "bool", 0, 0, 0 }
 };
 
 static Bool
@@ -2662,6 +2988,7 @@ blurInitScreen (CompPlugin *p,
 	bs->texture[i] = 0;
 
     bs->program   = 0;
+    bs->maxTemp   = 32;
     bs->fbo	  = 0;
     bs->fboStatus = FALSE;
 
@@ -2676,6 +3003,15 @@ blurInitScreen (CompPlugin *p,
     else
 	bs->alphaBlur = FALSE;
 
+    if (s->fragmentProgram)
+    {
+	int tmp[4];
+	s->getProgramiv (GL_FRAGMENT_PROGRAM_ARB,
+			 GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB,
+			 tmp);
+	bs->maxTemp = tmp[0];
+    }
+
     WRAP (bs, s, preparePaintScreen, blurPreparePaintScreen);
     WRAP (bs, s, donePaintScreen, blurDonePaintScreen);
     WRAP (bs, s, paintOutput, blurPaintOutput);
