@@ -287,8 +287,6 @@ dbusHandleRootIntrospectMessage (DBusConnection *connection,
 	return FALSE;
     }
 
-    dbusIntrospectAddNode (writer, "core");
-
     dbusIntrospectEndRoot (writer);
 
     xmlFreeTextWriter (writer);
@@ -551,9 +549,9 @@ dbusHandleOptionIntrospectMessage (DBusConnection *connection,
 		    strcpy (type, "as");
 		else
 		    strcpy (type, "s");
-
-	    default:
 		break;
+	    default:
+		continue;
 	    }
 
 	    dbusIntrospectAddMethod (writer,
@@ -671,6 +669,9 @@ dbusHandleActionMessage (DBusConnection *connection,
 	    while (object && object->type != COMP_OBJECT_TYPE_DISPLAY)
 		object = object->parent;
 
+	    if (!object)
+		return FALSE;
+
 	    if (activate)
 	    {
 		if (!option->value.action.initiate)
@@ -890,6 +891,9 @@ dbusGetOptionValue (CompObject	    *object,
 	    while (object && object->type != COMP_OBJECT_TYPE_DISPLAY)
 		object = object->parent;
 
+	    if (!object)
+		return FALSE;
+
 	    stringToKeyAction (GET_CORE_DISPLAY (object), s, &value->action);
 	    return TRUE;
 	}
@@ -902,6 +906,9 @@ dbusGetOptionValue (CompObject	    *object,
 	    while (object && object->type != COMP_OBJECT_TYPE_DISPLAY)
 		object = object->parent;
 
+	    if (!object)
+		return FALSE;
+
 	    stringToButtonAction (GET_CORE_DISPLAY (object),
 				  s, &value->action);
 	    return TRUE;
