@@ -29,7 +29,7 @@
 #include <string.h>
 #include <math.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 static CompMetadata minMetadata;
 
@@ -81,20 +81,20 @@ typedef struct _MinWindow {
     int unmapCnt;
 } MinWindow;
 
-#define GET_MIN_DISPLAY(d)				    \
-    ((MinDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_MIN_DISPLAY(d)					 \
+    ((MinDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define MIN_DISPLAY(d)			 \
     MinDisplay *md = GET_MIN_DISPLAY (d)
 
-#define GET_MIN_SCREEN(s, md)					\
-    ((MinScreen *) (s)->privates[(md)->screenPrivateIndex].ptr)
+#define GET_MIN_SCREEN(s, md)					     \
+    ((MinScreen *) (s)->base.privates[(md)->screenPrivateIndex].ptr)
 
 #define MIN_SCREEN(s)						     \
     MinScreen *ms = GET_MIN_SCREEN (s, GET_MIN_DISPLAY (s->display))
 
-#define GET_MIN_WINDOW(w, ms)				        \
-    ((MinWindow *) (w)->privates[(ms)->windowPrivateIndex].ptr)
+#define GET_MIN_WINDOW(w, ms)					     \
+    ((MinWindow *) (w)->base.privates[(ms)->windowPrivateIndex].ptr)
 
 #define MIN_WINDOW(w)					   \
     MinWindow *mw = GET_MIN_WINDOW  (w,			   \
@@ -803,6 +803,9 @@ minInitDisplay (CompPlugin  *p,
 {
     MinDisplay *md;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     md = malloc (sizeof (MinDisplay));
     if (!md)
 	return FALSE;
@@ -820,7 +823,7 @@ minInitDisplay (CompPlugin  *p,
 
     WRAP (md, d, handleEvent, minHandleEvent);
 
-    d->privates[displayPrivateIndex].ptr = md;
+    d->base.privates[displayPrivateIndex].ptr = md;
 
     return TRUE;
 }
@@ -886,7 +889,7 @@ minInitScreen (CompPlugin *p,
     WRAP (ms, s, damageWindowRect, minDamageWindowRect);
     WRAP (ms, s, focusWindow, minFocusWindow);
 
-    s->privates[md->screenPrivateIndex].ptr = ms;
+    s->base.privates[md->screenPrivateIndex].ptr = ms;
 
     return TRUE;
 }
@@ -953,7 +956,7 @@ minInitWindow (CompPlugin *p,
 	    mw->region = NULL;
     }
 
-    w->privates[ms->windowPrivateIndex].ptr = mw;
+    w->base.privates[ms->windowPrivateIndex].ptr = mw;
 
     return TRUE;
 }
@@ -973,6 +976,65 @@ minFiniWindow (CompPlugin *p,
     free (mw);
 }
 
+static CompBool
+minInitObject (CompPlugin *p,
+	       CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) minInitDisplay,
+	(InitPluginObjectProc) minInitScreen,
+	(InitPluginObjectProc) minInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+minFiniObject (CompPlugin *p,
+	       CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) minFiniDisplay,
+	(FiniPluginObjectProc) minFiniScreen,
+	(FiniPluginObjectProc) minFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+minGetObjectOptions (CompPlugin *plugin,
+		     CompObject *object,
+		     int	*count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) 0, /* GetDisplayOptions */
+	(GetPluginObjectOptionsProc) minGetScreenOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+minSetObjectOption (CompPlugin      *plugin,
+		    CompObject      *object,
+		    const char      *name,
+		    CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) 0, /* SetDisplayOption */
+	(SetPluginObjectOptionProc) minSetScreenOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 minInit (CompPlugin *p)
 {
@@ -1001,13 +1063,6 @@ minFini (CompPlugin *p)
     compFiniMetadata (&minMetadata);
 }
 
-static int
-minGetVersion (CompPlugin *plugin,
-	       int	  version)
-{
-    return ABIVERSION;
-}
-
 static CompMetadata *
 minGetMetadata (CompPlugin *plugin)
 {
@@ -1016,24 +1071,17 @@ minGetMetadata (CompPlugin *plugin)
 
 static CompPluginVTable minVTable = {
     "minimize",
-    minGetVersion,
     minGetMetadata,
     minInit,
     minFini,
-    minInitDisplay,
-    minFiniDisplay,
-    minInitScreen,
-    minFiniScreen,
-    minInitWindow,
-    minFiniWindow,
-    0, /* GetDisplayOptions */
-    0, /* SetDisplayOption */
-    minGetScreenOptions,
-    minSetScreenOption
+    minInitObject,
+    minFiniObject,
+    minGetObjectOptions,
+    minSetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &minVTable;
 }
