@@ -76,6 +76,8 @@ typedef struct _MoveDisplay {
     int        status;
     KeyCode    key[NUM_KEYS];
 
+    int releaseButton;
+
     GLushort moveOpacity;
 } MoveDisplay;
 
@@ -125,7 +127,7 @@ moveInitiate (CompDisplay     *d,
     {
 	XRectangle   workArea;
 	unsigned int mods;
-	int          x, y;
+	int          x, y, button;
 
 	MOVE_SCREEN (w->screen);
 
@@ -136,6 +138,8 @@ moveInitiate (CompDisplay     *d,
 	y = getIntOptionNamed (option, nOption, "y",
 			       w->attrib.y + (w->height / 2));
 
+	button = getIntOptionNamed (option, nOption, "button", -1);
+
 	if (otherScreenGrabExist (w->screen, "move", 0))
 	    return FALSE;
 
@@ -186,6 +190,8 @@ moveInitiate (CompDisplay     *d,
 	{
 	    md->w = w;
 
+	    md->releaseButton = button;
+
 	    (w->screen->windowGrabNotify) (w, x, y, mods,
 					   CompWindowGrabMoveMask |
 					   CompWindowGrabButtonMask);
@@ -245,7 +251,8 @@ moveTerminate (CompDisplay     *d,
 	if (md->moveOpacity != OPAQUE)
 	    addWindowDamage (md->w);
 
-	md->w = 0;
+	md->w             = 0;
+	md->releaseButton = 0;
     }
 
     action->state &= ~(CompActionStateTermKey | CompActionStateTermButton);
@@ -569,23 +576,26 @@ moveHandleEvent (CompDisplay *d,
 
 	    if (ms->grabIndex)
 	    {
-		CompAction *action;
-		int        opt = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
+		if (md->releaseButton == -1 ||
+		    md->releaseButton == event->xbutton.button)
+		{
+		    CompAction *action;
+		    int        opt = MOVE_DISPLAY_OPTION_INITIATE_BUTTON;
 
-		action = &md->opt[opt].value.action;
-		if (action->state & CompActionStateTermButton)
-		    moveTerminate (d, action, 0, NULL, 0);
+		    action = &md->opt[opt].value.action;
+		    moveTerminate (d, action, CompActionStateTermButton,
+				   NULL, 0);
+		}
 	    }
 	}
 	break;
     case KeyPress:
-    case KeyRelease:
 	s = findScreenAtDisplay (d, event->xkey.root);
 	if (s)
 	{
 	    MOVE_SCREEN (s);
 
-	    if (ms->grabIndex && event->type == KeyPress)
+	    if (ms->grabIndex)
 	    {
 		int i;
 
@@ -624,7 +634,7 @@ moveHandleEvent (CompDisplay *d,
 		w = findWindowAtDisplay (d, event->xclient.window);
 		if (w)
 		{
-		    CompOption o[4];
+		    CompOption o[5];
 		    int	       xRoot, yRoot;
 		    int	       option;
 
@@ -667,10 +677,15 @@ moveHandleEvent (CompDisplay *d,
 			    o[3].name	 = "y";
 			    o[3].value.i = event->xclient.data.l[1];
 
+			    o[4].type    = CompOptionTypeInt;
+			    o[4].name    = "button";
+			    o[4].value.i = event->xclient.data.l[3] ?
+				           event->xclient.data.l[3] : -1;
+
 			    moveInitiate (d,
 					  &md->opt[option].value.action,
 					  CompActionStateInitButton,
-					  o, 4);
+					  o, 5);
 
 			    moveHandleMotionEvent (w->screen, xRoot, yRoot);
 			}
@@ -836,9 +851,10 @@ moveInitDisplay (CompPlugin  *p,
     md->moveOpacity =
 	(md->opt[MOVE_DISPLAY_OPTION_OPACITY].value.i * OPAQUE) / 100;
 
-    md->w      = 0;
-    md->region = NULL;
-    md->status = RectangleOut;
+    md->w             = 0;
+    md->region        = NULL;
+    md->status        = RectangleOut;
+    md->releaseButton = 0;
 
     for (i = 0; i < NUM_KEYS; i++)
 	md->key[i] = XKeysymToKeycode (d->display,
