@@ -106,10 +106,10 @@ typedef struct _RotateDisplay {
 typedef struct _RotateScreen {
     PreparePaintScreenProc	 preparePaintScreen;
     DonePaintScreenProc		 donePaintScreen;
-    PaintScreenProc		 paintScreen;
     PaintOutputProc		 paintOutput;
     WindowGrabNotifyProc	 windowGrabNotify;
     WindowUngrabNotifyProc	 windowUngrabNotify;
+    ActivateWindowProc           activateWindow;
 
     CubeGetRotationProc getRotation;
 
@@ -141,8 +141,10 @@ typedef struct _RotateScreen {
     unsigned int      grabMask;
     CompWindow	      *grabWindow;
 
+    float progress;
+    float progressVelocity;
+
     GLfloat zoomTranslate;
-    GLfloat zoomVelocity;
 } RotateScreen;
 
 #define GET_ROTATE_DISPLAY(d)					    \
@@ -427,33 +429,27 @@ rotatePreparePaintScreen (CompScreen *s,
 	}
     }
 
-    if (rs->moving && cs->invert == 1 && !cs->unfolded)
+    if (rs->moving)
     {
 	if (fabs (rs->xrot + rs->baseXrot + rs->moveTo) <=
 	    (360.0 / (s->hsize * 2.0)))
 	{
-	    rs->zoomTranslate = rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f *
-		fabs (rs->xrot + rs->baseXrot + rs->moveTo) /
-		(360.0 / (s->hsize * 2.0));
+	    rs->progress = fabs (rs->xrot + rs->baseXrot + rs->moveTo) /
+			   (360.0 / (s->hsize * 2.0));
 	}
 	else if (fabs (rs->xrot + rs->baseXrot) <= (360.0 / (s->hsize * 2.0)))
 	{
-	    rs->zoomTranslate = rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f *
-		fabs (rs->xrot + rs->baseXrot) /
-		(360.0 / (s->hsize * 2.0));
+	    rs->progress = fabs (rs->xrot + rs->baseXrot) /
+			   (360.0 / (s->hsize * 2.0));
 	}
 	else
 	{
-	    rs->zoomTranslate += fabs (rs->xrot + rs->baseXrot - oldXrot) /
-		(360.0 / (s->hsize * 2.0)) *
-		rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f;
-	    rs->zoomTranslate =
-		MIN (rs->zoomTranslate,
-		     rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f);
+	    rs->progress += fabs (rs->xrot + rs->baseXrot - oldXrot) /
+			    (360.0 / (s->hsize * 2.0));
+	    rs->progress = MIN (rs->progress, 1.0);
 	}
     }
-    else if ((rs->zoomTranslate != 0.0f || rs->grabbed) && cs->invert == 1 &&
-	     !cs->unfolded)
+    else if (rs->progress != 0.0f || rs->grabbed)
     {
 	int   steps;
 	float amount, chunk;
@@ -472,10 +468,9 @@ rotatePreparePaintScreen (CompScreen *s,
 	    float dt, adjust, tamount;
 
 	    if (rs->grabbed)
-		dt = rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f -
-		     rs->zoomTranslate;
+		dt = 1.0 - rs->progress;
 	    else
-		dt = 0.0f - rs->zoomTranslate;
+		dt = 0.0f - rs->progress;
 
 	    adjust = dt * 0.15f;
 	    tamount = fabs (dt) * 1.5f;
@@ -484,23 +479,33 @@ rotatePreparePaintScreen (CompScreen *s,
 	    else if (tamount > 2.0f)
 		tamount = 2.0f;
 
-	    rs->zoomVelocity = (tamount * rs->zoomVelocity + adjust) /
-			       (tamount + 1.0f);
+	    rs->progressVelocity = (tamount * rs->progressVelocity + adjust) /
+				   (tamount + 1.0f);
 
-	    if (fabs (dt) < 0.1f && fabs (rs->zoomVelocity) < 0.0005f)
+	    rs->progress += rs->progressVelocity * chunk;
+
+	    if (fabs (dt) < 0.01f && fabs (rs->progressVelocity) < 0.0001f)
 	    {
 		if (rs->grabbed)
-		    rs->zoomTranslate =
-			rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f;
+		    rs->progress = 1.0f;
 		else
-		    rs->zoomTranslate = 0.0f;
+		    rs->progress = 0.0f;
 
 		break;
 	    }
-	    rs->zoomTranslate += rs->zoomVelocity * chunk;
 	}
     }
 
+    if (cs->invert == 1 && !cs->unfolded)
+    {
+	rs->zoomTranslate = rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f *
+			    rs->progress;
+    }
+    else
+    {
+	rs->zoomTranslate = 0.0;
+    }
+
     UNWRAP (rs, s, preparePaintScreen);
     (*s->preparePaintScreen) (s, msSinceLastPaint);
     WRAP (rs, s, preparePaintScreen, rotatePreparePaintScreen);
@@ -511,19 +516,16 @@ rotateDonePaintScreen (CompScreen *s)
 {
     ROTATE_SCREEN (s);
 
-    if (rs->grabIndex || rs->moving)
+    if (rs->grabIndex || rs->moving ||
+	(rs->progress != 0.0 && rs->progress != 1.0))
     {
 	if ((!rs->grabbed && !rs->snapTop && !rs->snapBottom) ||
-	    rs->xVelocity || rs->yVelocity)
+	    rs->xVelocity || rs->yVelocity || rs->progressVelocity)
 	{
 	    damageScreen (s);
 	}
     }
 
-    if (rs->zoomTranslate > 0.0f &&
-	rs->zoomTranslate < rs->opt[ROTATE_SCREEN_OPTION_ZOOM].value.f)
-	damageScreen (s);
-
     UNWRAP (rs, s, donePaintScreen);
     (*s->donePaintScreen) (s);
     WRAP (rs, s, donePaintScreen, rotateDonePaintScreen);
@@ -532,35 +534,19 @@ rotateDonePaintScreen (CompScreen *s)
 static void
 rotateGetRotation (CompScreen *s,
 		   float      *x,
-		   float      *v)
+		   float      *v,
+		   float      *progress)
 {
     CUBE_SCREEN (s);
     ROTATE_SCREEN (s);
 
     UNWRAP (rs, cs, getRotation);
-    (*cs->getRotation) (s, x, v);
+    (*cs->getRotation) (s, x, v, progress);
     WRAP (rs, cs, getRotation, rotateGetRotation);
 
     *x += rs->baseXrot + rs->xrot;
     *v += rs->yrot;
-}
-
-static void
-rotatePaintScreen (CompScreen   *s,
-		   CompOutput   *outputs,
-		   int          numOutputs,
-		   unsigned int mask)
-{
-    ROTATE_SCREEN (s);
-    CUBE_SCREEN (s);
-
-    UNWRAP (rs, s, paintScreen);
-    if ((rs->grabIndex || rs->moving || rs->zoomTranslate || cs->grabIndex) &&
-	cs->moMode == CUBE_MOMODE_ONE && s->nOutputDev)
-	(*s->paintScreen) (s, &s->fullscreenOutput, 1, mask);
-    else
-	(*s->paintScreen) (s, outputs, numOutputs, mask);
-    WRAP (rs, s, paintScreen, rotatePaintScreen);
+    *progress = MAX (*progress, rs->progress);
 }
 
 static Bool
@@ -575,7 +561,7 @@ rotatePaintOutput (CompScreen		   *s,
 
     ROTATE_SCREEN (s);
 
-    if (rs->grabIndex || rs->moving || rs->zoomTranslate != 0.0f)
+    if (rs->grabIndex || rs->moving || rs->progress != 0.0f)
     {
 	CompTransform sTransform = *transform;
 
@@ -1472,7 +1458,6 @@ rotateHandleEvent (CompDisplay *d,
 		   XEvent      *event)
 {
     CompScreen *s;
-    Window     activeWindow = d->activeWindow;
 
     ROTATE_DISPLAY (d);
 
@@ -1579,63 +1564,63 @@ rotateHandleEvent (CompDisplay *d,
     UNWRAP (rd, d, handleEvent);
     (*d->handleEvent) (d, event);
     WRAP (rd, d, handleEvent, rotateHandleEvent);
+}
 
-    if (activeWindow != d->activeWindow)
+static void
+rotateActivateWindow (CompWindow *w)
+{
+    CompScreen *s = w->screen;
+
+    ROTATE_SCREEN (s);
+
+    if (w->placed &&
+	!otherScreenGrabExist (s, "rotate", "switcher", "cube", 0))
     {
-	CompWindow *w;
+	int dx;
 
-	w = findWindowAtDisplay (d, d->activeWindow);
-	if (w && w->placed)
+	/* reset movement */
+	rs->moveTo = 0.0f;
+
+	defaultViewportForWindow (w, &dx, NULL);
+	dx -= s->x;
+	if (dx)
 	{
-	    s = w->screen;
+	    Window	 win;
+	    int		 i, x, y;
+	    unsigned int ui;
+	    CompOption   o[4];
 
-	    if (!otherScreenGrabExist (s, "rotate", "switcher", "cube", 0))
-	    {
-		int dx;
+	    XQueryPointer (s->display->display, s->root,
+			   &win, &win, &x, &y, &i, &i, &ui);
 
-		ROTATE_SCREEN (s);
+	    if (dx * 2 > s->hsize)
+		dx -= s->hsize;
+	    else if (dx * 2 < -s->hsize)
+		dx += s->hsize;
 
-		/* reset movement */
-		rs->moveTo = 0.0f;
+	    o[0].type    = CompOptionTypeInt;
+	    o[0].name    = "x";
+	    o[0].value.i = x;
 
-		defaultViewportForWindow (w, &dx, NULL);
-		dx -= s->x;
-		if (dx)
-		{
-		    Window	 win;
-		    int		 i, x, y;
-		    unsigned int ui;
-		    CompOption   o[4];
+	    o[1].type    = CompOptionTypeInt;
+	    o[1].name    = "y";
+	    o[1].value.i = y;
 
-		    XQueryPointer (d->display, s->root,
-				   &win, &win, &x, &y, &i, &i, &ui);
+	    o[2].type    = CompOptionTypeInt;
+	    o[2].name    = "root";
+	    o[2].value.i = s->root;
 
-		    if (dx * 2 > s->hsize)
-			dx -= s->hsize;
-		    else if (dx * 2 < -s->hsize)
-			dx += s->hsize;
+	    o[3].type    = CompOptionTypeInt;
+	    o[3].name    = "direction";
+	    o[3].value.i = dx;
 
-		    o[0].type    = CompOptionTypeInt;
-		    o[0].name    = "x";
-		    o[0].value.i = x;
-
-		    o[1].type    = CompOptionTypeInt;
-		    o[1].name    = "y";
-		    o[1].value.i = y;
-
-		    o[2].type	 = CompOptionTypeInt;
-		    o[2].name	 = "root";
-		    o[2].value.i = s->root;
-
-		    o[3].type	 = CompOptionTypeInt;
-		    o[3].name	 = "direction";
-		    o[3].value.i = dx;
-
-		    rotate (d, NULL, 0, o, 4);
-		}
-	    }
+	    rotate (s->display, NULL, 0, o, 4);
 	}
     }
+
+    UNWRAP (rs, s, activateWindow);
+    (*s->activateWindow) (w);
+    WRAP (rs, s, activateWindow, rotateActivateWindow);
 }
 
 static void
@@ -1871,15 +1856,17 @@ rotateInitScreen (CompPlugin *p,
 
     rs->rotateHandle = 0;
 
+    rs->progress          = 0.0;
+    rs->progressVelocity  = 0.0;
+
     rs->zoomTranslate = 0.0;
-    rs->zoomVelocity  = 0.0;
 
     WRAP (rs, s, preparePaintScreen, rotatePreparePaintScreen);
     WRAP (rs, s, donePaintScreen, rotateDonePaintScreen);
-    WRAP (rs, s, paintScreen, rotatePaintScreen);
     WRAP (rs, s, paintOutput, rotatePaintOutput);
     WRAP (rs, s, windowGrabNotify, rotateWindowGrabNotify);
     WRAP (rs, s, windowUngrabNotify, rotateWindowUngrabNotify);
+    WRAP (rs, s, activateWindow, rotateActivateWindow);
 
     WRAP (rs, cs, getRotation, rotateGetRotation);
 
@@ -1899,10 +1886,10 @@ rotateFiniScreen (CompPlugin *p,
 
     UNWRAP (rs, s, preparePaintScreen);
     UNWRAP (rs, s, donePaintScreen);
-    UNWRAP (rs, s, paintScreen);
     UNWRAP (rs, s, paintOutput);
     UNWRAP (rs, s, windowGrabNotify);
     UNWRAP (rs, s, windowUngrabNotify);
+    UNWRAP (rs, s, activateWindow);
 
     compFiniScreenOptions (s, rs->opt, ROTATE_SCREEN_OPTION_NUM);
 
