@@ -986,33 +986,6 @@ scaleCheckForWindowAt (CompScreen *s,
     return 0;
 }
 
-static void
-sendViewportMoveRequest (CompScreen *s,
-			 int	    x,
-			 int	    y)
-{
-    XEvent xev;
-
-    xev.xclient.type    = ClientMessage;
-    xev.xclient.display = s->display->display;
-    xev.xclient.format  = 32;
-
-    xev.xclient.message_type = s->display->desktopViewportAtom;
-    xev.xclient.window	     = s->root;
-
-    xev.xclient.data.l[0] = x;
-    xev.xclient.data.l[1] = y;
-    xev.xclient.data.l[2] = 0;
-    xev.xclient.data.l[3] = 0;
-    xev.xclient.data.l[4] = 0;
-
-    XSendEvent (s->display->display,
-		s->root,
-		FALSE,
-		SubstructureRedirectMask | SubstructureNotifyMask,
-		&xev);
-}
-
 static void
 sendDndStatusMessage (CompScreen *s,
 		      Window	 source)
@@ -1097,20 +1070,9 @@ scaleTerminate (CompDisplay     *d,
 		}
 		else if (ss->state != SCALE_STATE_IN)
 		{
-		    w = findWindowAtScreen (s, sd->lastActiveWindow);
+		    w = findWindowAtScreen (s, sd->selectedWindow);
 		    if (w)
-		    {
-			int x, y;
-
-			activateWindow (w);
-
-			defaultViewportForWindow (w, &x, &y);
-
-			if (x != s->x || y != s->y)
-			    sendViewportMoveRequest (s,
-						     x * s->width,
-						     y * s->height);
-		    }
+			(*s->activateWindow) (w);
 		}
 
 		ss->state = SCALE_STATE_IN;
@@ -1221,8 +1183,11 @@ scaleInitiateCommon (CompScreen      *s,
 	damageScreen (s);
     }
 
-    if (state & CompActionStateInitButton)
+    if ((state & (CompActionStateInitButton | EDGE_STATE)) ==
+	CompActionStateInitButton)
+    {
 	action->state |= CompActionStateTermButton;
+    }
 
     if (state & CompActionStateInitKey)
 	action->state |= CompActionStateTermKey;
