@@ -32,7 +32,7 @@
 #include <string.h>
 #include <math.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 
 #define TEXTURE_SIZE 256
 
@@ -66,15 +66,15 @@ static int displayPrivateIndex;
 static int waterLastPointerX = 0;
 static int waterLastPointerY = 0;
 
-#define WATER_DISPLAY_OPTION_INITIATE     0
-#define WATER_DISPLAY_OPTION_TOGGLE_RAIN  1
-#define WATER_DISPLAY_OPTION_TOGGLE_WIPER 2
-#define WATER_DISPLAY_OPTION_OFFSET_SCALE 3
-#define WATER_DISPLAY_OPTION_RAIN_DELAY	  4
-#define WATER_DISPLAY_OPTION_TITLE_WAVE   5
-#define WATER_DISPLAY_OPTION_POINT        6
-#define WATER_DISPLAY_OPTION_LINE         7
-#define WATER_DISPLAY_OPTION_NUM          8
+#define WATER_DISPLAY_OPTION_INITIATE_KEY     0
+#define WATER_DISPLAY_OPTION_TOGGLE_RAIN_KEY  1
+#define WATER_DISPLAY_OPTION_TOGGLE_WIPER_KEY 2
+#define WATER_DISPLAY_OPTION_OFFSET_SCALE     3
+#define WATER_DISPLAY_OPTION_RAIN_DELAY	      4
+#define WATER_DISPLAY_OPTION_TITLE_WAVE       5
+#define WATER_DISPLAY_OPTION_POINT            6
+#define WATER_DISPLAY_OPTION_LINE             7
+#define WATER_DISPLAY_OPTION_NUM              8
 
 typedef struct _WaterDisplay {
     int		    screenPrivateIndex;
@@ -120,14 +120,14 @@ typedef struct _WaterScreen {
     WaterFunction *bumpMapFunctions;
 } WaterScreen;
 
-#define GET_WATER_DISPLAY(d)				      \
-    ((WaterDisplay *) (d)->privates[displayPrivateIndex].ptr)
+#define GET_WATER_DISPLAY(d)					   \
+    ((WaterDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define WATER_DISPLAY(d)		     \
     WaterDisplay *wd = GET_WATER_DISPLAY (d)
 
-#define GET_WATER_SCREEN(s, wd)					  \
-    ((WaterScreen *) (s)->privates[(wd)->screenPrivateIndex].ptr)
+#define GET_WATER_SCREEN(s, wd)					       \
+    ((WaterScreen *) (s)->base.privates[(wd)->screenPrivateIndex].ptr)
 
 #define WATER_SCREEN(s)							   \
     WaterScreen *ws = GET_WATER_SCREEN (s, GET_WATER_DISPLAY (s->display))
@@ -1532,7 +1532,7 @@ waterGetDisplayOptions (CompPlugin  *plugin,
 static Bool
 waterSetDisplayOption (CompPlugin      *plugin,
 		       CompDisplay     *display,
-		       char	       *name,
+		       const char      *name,
 		       CompOptionValue *value)
 {
     CompOption *o;
@@ -1578,12 +1578,12 @@ waterSetDisplayOption (CompPlugin      *plugin,
 }
 
 static const CompMetadataOptionInfo waterDisplayOptionInfo[] = {
-    { "initiate", "action", 0, waterInitiate, waterTerminate },
-    { "toggle_rain", "action", 0, waterToggleRain, 0 },
-    { "toggle_wiper", "action", 0, waterToggleWiper, 0 },
+    { "initiate_key", "key", 0, waterInitiate, waterTerminate },
+    { "toggle_rain_key", "key", 0, waterToggleRain, 0 },
+    { "toggle_wiper_key", "key", 0, waterToggleWiper, 0 },
     { "offset_scale", "float", "<min>0</min>", 0, 0 },
     { "rain_delay", "int", "<min>1</min>", 0, 0 },
-    { "title_wave", "action", 0, waterTitleWave, 0 },
+    { "title_wave", "bell", 0, waterTitleWave, 0 },
     { "point", "action", 0, waterPoint, 0 },
     { "line", "action", 0, waterLine, 0 }
 };
@@ -1594,6 +1594,9 @@ waterInitDisplay (CompPlugin  *p,
 {
     WaterDisplay *wd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     wd = malloc (sizeof (WaterDisplay));
     if (!wd)
 	return FALSE;
@@ -1620,7 +1623,7 @@ waterInitDisplay (CompPlugin  *p,
 
     WRAP (wd, d, handleEvent, waterHandleEvent);
 
-    d->privates[displayPrivateIndex].ptr = wd;
+    d->base.privates[displayPrivateIndex].ptr = wd;
 
     return TRUE;
 }
@@ -1658,7 +1661,7 @@ waterInitScreen (CompPlugin *p,
     WRAP (ws, s, donePaintScreen, waterDonePaintScreen);
     WRAP (ws, s, drawWindowTexture, waterDrawWindowTexture);
 
-    s->privates[wd->screenPrivateIndex].ptr = ws;
+    s->base.privates[wd->screenPrivateIndex].ptr = ws;
 
     waterReset (s);
 
@@ -1712,6 +1715,61 @@ waterFiniScreen (CompPlugin *p,
     free (ws);
 }
 
+static CompBool
+waterInitObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) waterInitDisplay,
+	(InitPluginObjectProc) waterInitScreen
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+waterFiniObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) waterFiniDisplay,
+	(FiniPluginObjectProc) waterFiniScreen
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
+static CompOption *
+waterGetObjectOptions (CompPlugin *plugin,
+		       CompObject *object,
+		       int	  *count)
+{
+    static GetPluginObjectOptionsProc dispTab[] = {
+	(GetPluginObjectOptionsProc) 0, /* GetCoreOptions */
+	(GetPluginObjectOptionsProc) waterGetDisplayOptions
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
+		     (void *) (*count = 0), (plugin, object, count));
+}
+
+static CompBool
+waterSetObjectOption (CompPlugin      *plugin,
+		      CompObject      *object,
+		      const char      *name,
+		      CompOptionValue *value)
+{
+    static SetPluginObjectOptionProc dispTab[] = {
+	(SetPluginObjectOptionProc) 0, /* SetCoreOption */
+	(SetPluginObjectOptionProc) waterSetDisplayOption
+    };
+
+    RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab), FALSE,
+		     (plugin, object, name, value));
+}
+
 static Bool
 waterInit (CompPlugin *p)
 {
@@ -1741,13 +1799,6 @@ waterFini (CompPlugin *p)
     compFiniMetadata (&waterMetadata);
 }
 
-static int
-waterGetVersion (CompPlugin *plugin,
-		 int	    version)
-{
-    return ABIVERSION;
-}
-
 static CompMetadata *
 waterGetMetadata (CompPlugin *plugin)
 {
@@ -1756,24 +1807,17 @@ waterGetMetadata (CompPlugin *plugin)
 
 static CompPluginVTable waterVTable = {
     "water",
-    waterGetVersion,
     waterGetMetadata,
     waterInit,
     waterFini,
-    waterInitDisplay,
-    waterFiniDisplay,
-    waterInitScreen,
-    waterFiniScreen,
-    0, /* InitWindow */
-    0, /* FiniWindow */
-    waterGetDisplayOptions,
-    waterSetDisplayOption,
-    0, /* GetScreenOptions */
-    0  /* SetScreenOption */
+    waterInitObject,
+    waterFiniObject,
+    waterGetObjectOptions,
+    waterSetObjectOption
 };
 
 CompPluginVTable *
-getCompPluginInfo (void)
+getCompPluginInfo20070830 (void)
 {
     return &waterVTable;
 }
