@@ -173,7 +173,7 @@ closeWin (CompDisplay     *d,
     time = getIntOptionNamed (option, nOption, "time", CurrentTime);
 
     w = findTopLevelWindowAtDisplay (d, xid);
-    if (w)
+    if (w && (w->actions & CompWindowActionCloseMask))
 	closeWindow (w, time);
 
     return TRUE;
@@ -255,7 +255,7 @@ minimize (CompDisplay     *d,
     xid = getIntOptionNamed (option, nOption, "window", 0);
 
     w = findTopLevelWindowAtDisplay (d, xid);
-    if (w)
+    if (w && (w->actions & CompWindowActionMinimizeMask))
 	minimizeWindow (w);
 
     return TRUE;
@@ -577,7 +577,7 @@ windowMenu (CompDisplay     *d,
     xid = getIntOptionNamed (option, nOption, "window", 0);
 
     w = findTopLevelWindowAtDisplay (d, xid);
-    if (w)
+    if (w && !w->screen->maxGrab)
     {
 	int  x, y, button;
 	Time time;
@@ -753,7 +753,8 @@ const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "ignore_hints_when_maximized", "bool", 0, 0, 0 },
     { "command_terminal", "string", 0, 0, 0 },
     { "run_command_terminal_key", "key", 0, runCommandTerminal, 0 },
-    { "ping_delay", "int", "<min>1000</min>", 0, 0 }
+    { "ping_delay", "int", "<min>1000</min>", 0, 0 },
+    { "edge_delay", "int", "<min>0</min>", 0, 0 }
 };
 
 CompOption *
@@ -924,17 +925,23 @@ updatePlugins (CompDisplay *d)
     d->dirtyPluginList = FALSE;
 
     o = &d->opt[COMP_DISPLAY_OPTION_ACTIVE_PLUGINS];
-    for (i = 0; i < d->plugin.list.nValue && i < o->value.list.nValue; i++)
+
+    /* The old plugin list always begins with the core plugin. To make sure
+       we don't unnecessarily unload plugins if the new plugin list does not
+       contain the core plugin, we have to use an offset */
+    if (o->value.list.nValue > 0 && strcmp (o->value.list.value[0].s, "core"))
+	i = 0;
+    else
+	i = 1;
+
+    /* j is initialized to 1 to make sure we never pop the core plugin */
+    for (j = 1; j < d->plugin.list.nValue && i < o->value.list.nValue; i++, j++)
     {
-	if (strcmp (d->plugin.list.value[i].s, o->value.list.value[i].s))
+	if (strcmp (d->plugin.list.value[j].s, o->value.list.value[i].s))
 	    break;
     }
 
-    /* never pop the core plugin */
-    if (i)
-	nPop = d->plugin.list.nValue - i;
-    else
-	nPop = d->plugin.list.nValue - 1;
+    nPop = d->plugin.list.nValue - j;
 
     if (nPop)
     {
@@ -1728,9 +1735,14 @@ eventLoop (void)
 				glClear (GL_COLOR_BUFFER_BIT);
 			}
 
-			(*s->paintScreen) (s, s->outputDev,
-					   s->nOutputDev,
-					   mask);
+			if (s->opt[COMP_SCREEN_OPTION_FORCE_INDEPENDENT].value.b
+			    || !s->hasOverlappingOutputs)
+			    (*s->paintScreen) (s, s->outputDev,
+					       s->nOutputDev,
+					       mask);
+			else
+			    (*s->paintScreen) (s, &s->fullscreenOutput, 1,
+					       mask);
 
 			targetScreen = NULL;
 			targetOutput = &s->outputDev[0];
@@ -1995,6 +2007,8 @@ addDisplay (const char *name)
     d->screenPrivateIndices = 0;
     d->screenPrivateLen     = 0;
 
+    d->edgeDelayHandle = 0;
+
     d->logMessage = logMessage;
 
     d->modMap = 0;
