@@ -604,6 +604,7 @@ isEdgeAction (CompOption      *option,
 static Bool
 isEdgeEnterAction (CompOption      *option,
 		   CompActionState state,
+		   CompActionState delayState,
 		   unsigned int    edge,
 		   CompAction      **action)
 {
@@ -616,6 +617,18 @@ isEdgeEnterAction (CompOption      *option,
     if (!option->value.action.initiate)
 	return FALSE;
 
+    if (delayState)
+    {
+	if ((option->value.action.state & CompActionStateNoEdgeDelay) !=
+	    (delayState & CompActionStateNoEdgeDelay))
+	{
+	    /* ignore edge actions which shouldn't be delayed when invoking
+	       undelayed edges (or vice versa) */
+	    return FALSE;
+	}
+    }
+
+
     *action = &option->value.action;
 
     return TRUE;
@@ -643,6 +656,7 @@ triggerEdgeEnterBindings (CompDisplay	  *d,
 			  CompOption	  *option,
 			  int		  nOption,
 			  CompActionState state,
+			  CompActionState delayState,
 			  unsigned int	  edge,
 			  CompOption	  *argument,
 			  int		  nArgument)
@@ -651,7 +665,7 @@ triggerEdgeEnterBindings (CompDisplay	  *d,
 
     while (nOption--)
     {
-	if (isEdgeEnterAction (option, state, edge, &action))
+	if (isEdgeEnterAction (option, state, delayState, edge, &action))
 	{
 	    if ((*action->initiate) (d, action, state, argument, nArgument))
 		return TRUE;
@@ -688,6 +702,106 @@ triggerEdgeLeaveBindings (CompDisplay	  *d,
     return FALSE;
 }
 
+static Bool
+triggerAllEdgeEnterBindings (CompDisplay     *d,
+			     CompActionState state,
+			     CompActionState delayState,
+			     unsigned int    edge,
+			     CompOption	     *argument,
+			     int	     nArgument)
+{
+    CompOption *option;
+    int        nOption;
+    CompPlugin *p;
+
+    for (p = getPlugins (); p; p = p->next)
+    {
+	if (p->vTable->getObjectOptions)
+	{
+	    option = (*p->vTable->getObjectOptions) (p, &d->base, &nOption);
+	    if (triggerEdgeEnterBindings (d,
+					  option, nOption,
+					  state, delayState, edge,
+					  argument, nArgument))
+	    {
+		return TRUE;
+	    }
+	}
+    }
+    return FALSE;
+}
+
+static Bool
+delayedEdgeTimeout (void *closure)
+{
+    CompDelayedEdgeSettings *settings = (CompDelayedEdgeSettings *) closure;
+
+    triggerAllEdgeEnterBindings (settings->d,
+				 settings->state,
+				 ~CompActionStateNoEdgeDelay,
+				 settings->edge,
+				 settings->option,
+				 settings->nOption);
+
+    free (settings);
+
+    return FALSE;
+}
+
+static Bool
+triggerEdgeEnter (CompDisplay     *d,
+		  unsigned int    edge,
+		  CompActionState state,
+		  CompOption      *argument,
+		  unsigned int    nArgument)
+{
+    int                     delay;
+    CompDelayedEdgeSettings *delayedSettings = NULL;
+
+    delay = d->opt[COMP_DISPLAY_OPTION_EDGE_DELAY].value.i;
+
+    if (nArgument > 7)
+	nArgument = 7;
+
+    if (delay > 0)
+    {
+	delayedSettings = malloc (sizeof (CompDelayedEdgeSettings));
+	if (delayedSettings)
+	{
+	    delayedSettings->d       = d;
+	    delayedSettings->edge    = edge;
+	    delayedSettings->state   = state;
+	    delayedSettings->nOption = nArgument;
+	}
+    }
+
+    if (delayedSettings)
+    {
+	CompActionState delayState;
+	int             i;
+
+	for (i = 0; i < nArgument; i++)
+	    delayedSettings->option[i] = argument[i];
+
+	d->edgeDelayHandle = compAddTimeout (delay,
+					     delayedEdgeTimeout,
+					     delayedSettings);
+
+	delayState = CompActionStateNoEdgeDelay;
+	if (triggerAllEdgeEnterBindings (d, state, delayState,
+					 edge, argument, nArgument))
+	    return TRUE;
+    }
+    else
+    {
+	if (triggerAllEdgeEnterBindings (d, state, 0, edge,
+					 argument, nArgument))
+	    return TRUE;
+    }
+
+    return FALSE;
+}
+
 static Bool
 handleActionEvent (CompDisplay *d,
 		   XEvent      *event)
@@ -833,6 +947,16 @@ handleActionEvent (CompDisplay *d,
 	    if (!s)
 		return FALSE;
 
+	    if (d->edgeDelayHandle)
+	    {
+		void *closure;
+
+		closure = compRemoveTimeout (d->edgeDelayHandle);
+		if (closure)
+		    free (closure);
+		d->edgeDelayHandle = 0;
+	    }
+
 	    if (edgeWindow && edgeWindow != event->xcrossing.window)
 	    {
 		state = CompActionStateTermEdge;
@@ -900,18 +1024,11 @@ handleActionEvent (CompDisplay *d,
 		o[6].name    = "time";
 		o[6].value.i = event->xcrossing.time;
 
-		for (p = getPlugins (); p; p = p->next)
-		{
-		    if (!p->vTable->getObjectOptions)
-			continue;
-
-		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
-		    if (triggerEdgeEnterBindings (d, option, nOption, state,
-						  edge, o, 7))
-			return TRUE;
-		}
+		if (triggerEdgeEnter (d, edge, state, o, 7))
+		    return TRUE;
 	    }
-	} break;
+	}
+	break;
     case ClientMessage:
 	if (event->xclient.message_type == d->xdndEnterAtom)
 	{
@@ -1007,16 +1124,8 @@ handleActionEvent (CompDisplay *d,
 		o[4].value.i = event->xclient.data.l[2] & 0xffff;
 		o[5].value.i = root;
 
-		for (p = getPlugins (); p; p = p->next)
-		{
-		    if (!p->vTable->getObjectOptions)
-			continue;
-
-		    option = (*p->vTable->getObjectOptions) (p, obj, &nOption);
-		    if (triggerEdgeEnterBindings (d, option, nOption, state,
-						  edge, o, 6))
-			return TRUE;
-		}
+		if (triggerEdgeEnter (d, edge, state, o, 6))
+		    return TRUE;
 	    }
 
 	    xdndWindow = None;
@@ -1146,18 +1255,10 @@ handleEvent (CompDisplay *d,
 
 	    if (s->nExpose == s->sizeExpose)
 	    {
-		if (s->exposeRects)
-		{
-		    s->exposeRects = realloc (s->exposeRects,
+		s->exposeRects = realloc (s->exposeRects,
 					      (s->sizeExpose + more) *
 					      sizeof (XRectangle));
-		    s->sizeExpose += more;
-		}
-		else
-		{
-		    s->exposeRects = malloc (more * sizeof (XRectangle));
-		    s->sizeExpose = more;
-		}
+		s->sizeExpose += more;
 	    }
 
 	    s->exposeRects[s->nExpose].x      = event->xexpose.x;
@@ -1231,6 +1332,9 @@ handleEvent (CompDisplay *d,
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
 	{
+	    if (!w->attrib.override_redirect)
+		w->managed = TRUE;
+
 	    /* been shaded */
 	    if (w->height == 0)
 	    {
@@ -1408,7 +1512,10 @@ handleEvent (CompDisplay *d,
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
+	    {
 		updateTransientHint (w);
+		recalcWindowActions (w);
+	    }
 	}
 	else if (event->xproperty.atom == d->wmClientLeaderAtom)
 	{
@@ -1549,7 +1656,7 @@ handleEvent (CompDisplay *d,
 		if (event->xclient.data.l[0] != 1 ||
 		    allowWindowFocus (w, 0, event->xclient.data.l[1]))
 		{
-		    activateWindow (w);
+		    (*w->screen->activateWindow) (w);
 		}
 	    }
 	}
@@ -1618,6 +1725,8 @@ handleEvent (CompDisplay *d,
 		}
 
 		wState = constrainWindowState (wState, w->actions);
+		if (w->id == d->activeWindow)
+		    wState &= ~CompWindowStateDemandsAttentionMask;
 
 		if (wState != w->state)
 		{
@@ -1637,9 +1746,6 @@ handleEvent (CompDisplay *d,
 
 		    changeWindowState (w, wState);
 
-		    recalcWindowType (w);
-		    recalcWindowActions (w);
-
 		    updateWindowAttributes (w, stackingUpdateMode);
 		}
 	    }
@@ -1772,7 +1878,10 @@ handleEvent (CompDisplay *d,
 	    if (w)
 	    {
 		if (event->xclient.data.l[0] == IconicState)
-		    minimizeWindow (w);
+		{
+		    if (w->actions & CompWindowActionMinimizeMask)
+			minimizeWindow (w);
+		}
 		else if (event->xclient.data.l[0] == NormalState)
 		    unminimizeWindow (w);
 	    }
@@ -1862,18 +1971,32 @@ handleEvent (CompDisplay *d,
 
 		if (!w->placed)
 		{
-		    int newX, newY;
+		    int            newX, newY;
+		    int            gravity = w->sizeHints.win_gravity;
+		    XWindowChanges xwc;
+		    unsigned int   xwcm;
 
-		    if ((*w->screen->placeWindow) (w, w->serverX, w->serverY,
+		    /* adjust for gravity */
+		    xwc.x      = w->serverX;
+		    xwc.y      = w->serverY;
+		    xwc.width  = w->serverWidth;
+		    xwc.height = w->serverHeight;
+
+		    xwcm = adjustConfigureRequestForGravity (w, &xwc,
+							     CWX | CWY,
+							     gravity);
+
+		    if ((*w->screen->placeWindow) (w, xwc.x, xwc.y,
 						   &newX, &newY))
 		    {
-			moveWindow (w,
-				    newX - w->attrib.x,
-				    newY - w->attrib.y,
-				    FALSE, TRUE);
-			syncWindowPosition (w);
+			xwc.x = newX;
+			xwc.y = newY;
+			xwcm |= CWX | CWY;
 		    }
 
+		    if (xwcm)
+			configureXWindow (w, xwcm, &xwc);
+
 		    w->placed   = TRUE;
 		}
 
@@ -2103,18 +2226,10 @@ handleEvent (CompDisplay *d,
 		{
 		    if (w->nDamage == w->sizeDamage)
 		    {
-			if (w->damageRects)
-			{
-			    w->damageRects = realloc (w->damageRects,
-						      (w->sizeDamage + 1) *
-						      sizeof (XRectangle));
-			    w->sizeDamage += 1;
-			}
-			else
-			{
-			    w->damageRects = malloc (sizeof (XRectangle));
-			    w->sizeDamage  = 1;
-			}
+			w->damageRects = realloc (w->damageRects,
+						  (w->sizeDamage + 1) *
+						  sizeof (XRectangle));
+			w->sizeDamage += 1;
 		    }
 
 		    w->damageRects[w->nDamage].x      = de->area.x;
