@@ -36,6 +36,7 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <assert.h>
+#include <limits.h>
 
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
@@ -266,7 +267,7 @@ updateOutputDevices (CompScreen	*s)
     CompOutput	  *o, *output = NULL;
     CompListValue *list = &s->opt[COMP_SCREEN_OPTION_OUTPUTS].value.list;
     int		  nOutput = 0;
-    int		  x, y, i, bits;
+    int		  x, y, i, j, bits;
     unsigned int  width, height;
     int		  x1, y1, x2, y2;
     Region	  region;
@@ -366,8 +367,9 @@ updateOutputDevices (CompScreen	*s)
 	free (s->outputDev);
     }
 
-    s->outputDev  = output;
-    s->nOutputDev = nOutput;
+    s->outputDev             = output;
+    s->nOutputDev            = nOutput;
+    s->hasOverlappingOutputs = FALSE;
 
     setCurrentOutput (s, s->currentOutputDev);
 
@@ -384,6 +386,17 @@ updateOutputDevices (CompScreen	*s)
 	r.rects = &r.extents;
 	r.numRects = 1;
 
+	for (i = 0; i < nOutput - 1; i++)
+	    for (j = i + 1; j < nOutput; j++)
+            {
+		XIntersectRegion (&output[i].region,
+				  &output[j].region,
+				  region);
+		if (REGION_NOT_EMPTY (region))
+		    s->hasOverlappingOutputs = TRUE;
+	    }
+	XSubtractRegion (&emptyRegion, &emptyRegion, region);
+	
 	if (s->display->nScreenInfo)
 	{
 	    for (i = 0; i < s->display->nScreenInfo; i++)
@@ -616,6 +629,13 @@ setScreenOption (CompPlugin	 *plugin,
 	    return TRUE;
 	}
 	break;
+     case COMP_SCREEN_OPTION_FORCE_INDEPENDENT:
+	if (compSetBoolOption (o, value))
+	{
+	    updateOutputDevices (screen);
+	    return TRUE;
+	}
+	break;
     default:
 	if (compSetScreenOption (screen, o, value))
 	    return TRUE;
@@ -638,9 +658,15 @@ const CompMetadataOptionInfo coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM] = {
     { "number_of_desktops", "int", "<min>1</min>", 0, 0 },
     { "detect_outputs", "bool", 0, 0, 0 },
     { "outputs", "list", "<type>string</type>", 0, 0 },
+    { "overlapping_outputs", "int",
+      RESTOSTRING (0, OUTPUT_OVERLAP_MODE_LAST), 0, 0 },
+    { "focus_prevention_level", "int",
+      RESTOSTRING (0, FOCUS_PREVENTION_LEVEL_LAST), 0, 0 },
     { "focus_prevention_match", "match", 0, 0, 0 },
     { "opacity_matches", "list", "<type>match</type>", 0, 0 },
-    { "opacity_values", "list", "<type>int</type>", 0, 0 }
+    { "opacity_values", "list", "<type>int</type>", 0, 0 },
+    { "texture_compression", "bool", 0, 0, 0 },
+    { "force_independent_output_painting", "bool", 0, 0, 0 }
 };
 
 static void
@@ -1344,6 +1370,50 @@ hideOutputWindow (CompScreen *s)
 
 }
 
+void
+updateOutputWindow (CompScreen *s)
+{
+
+#ifdef USE_COW
+    if (useCow)
+    {
+	Display       *dpy = s->display->display;
+	XserverRegion region;
+	static Region tmpRegion = NULL;
+	CompWindow    *w;
+
+	if (!tmpRegion)
+	{
+	    tmpRegion = XCreateRegion ();
+	    if (!tmpRegion)
+		return;
+	}
+
+	XSubtractRegion (&s->region, &emptyRegion, tmpRegion);
+
+	for (w = s->reverseWindows; w; w = w->prev)
+	    if (w->overlayWindow)
+	    {
+		XSubtractRegion (tmpRegion, w->region, tmpRegion);
+	    }
+	
+	XShapeCombineRegion (dpy, s->output, ShapeBounding,
+			     0, 0, tmpRegion, ShapeSet);
+
+
+	region = XFixesCreateRegion (dpy, NULL, 0);
+
+	XFixesSetWindowShapeRegion (dpy,
+				    s->output,
+				    ShapeInput,
+				    0, 0, region);
+
+	XFixesDestroyRegion (dpy, region);
+    }
+#endif
+
+}
+
 static void
 makeOutputWindow (CompScreen *s)
 {
@@ -1381,7 +1451,8 @@ enterShowDesktopMode (CompScreen *s)
 	if ((s->showingDesktopMask & w->wmType) &&
 	    (!(w->state & CompWindowStateSkipTaskbarMask) || st->value.b))
 	{
-	    if (!w->inShowDesktopMode && !w->grabbed && (*s->focusWindow) (w))
+	    if (!w->inShowDesktopMode && !w->grabbed &&
+		w->managed && (*s->focusWindow) (w))
 	    {
 		w->inShowDesktopMode = TRUE;
 		hideWindow (w);
@@ -1699,7 +1770,6 @@ addScreen (CompDisplay *display,
     s->enableOutputClipping	   = enableOutputClipping;
     s->disableOutputClipping	   = disableOutputClipping;
     s->applyScreenTransform	   = applyScreenTransform;
-    s->paintBackground		   = paintBackground;
     s->paintWindow		   = paintWindow;
     s->drawWindow		   = drawWindow;
     s->addWindowGeometry	   = addWindowGeometry;
@@ -1708,6 +1778,7 @@ addScreen (CompDisplay *display,
     s->getOutputExtentsForWindow   = getOutputExtentsForWindow;
     s->getAllowedActionsForWindow  = getAllowedActionsForWindow;
     s->focusWindow		   = focusWindow;
+    s->activateWindow              = activateWindow;
     s->placeWindow                 = placeWindow;
     s->validateWindowResizeRequest = validateWindowResizeRequest;
 
@@ -1953,8 +2024,10 @@ addScreen (CompDisplay *display,
 	    getProcAddress (s, "glActiveTexture");
 	s->clientActiveTexture = (GLClientActiveTextureProc)
 	    getProcAddress (s, "glClientActiveTexture");
+	s->multiTexCoord2f = (GLMultiTexCoord2fProc)
+	    getProcAddress (s, "glMultiTexCoord2f");
 
-	if (s->activeTexture && s->clientActiveTexture)
+	if (s->activeTexture && s->clientActiveTexture && s->multiTexCoord2f)
 	    glGetIntegerv (GL_MAX_TEXTURE_UNITS_ARB, &s->maxTextureUnits);
     }
 
@@ -1973,13 +2046,16 @@ addScreen (CompDisplay *display,
 	    getProcAddress (s, "glProgramEnvParameter4fARB");
 	s->programLocalParameter4f = (GLProgramParameter4fProc)
 	    getProcAddress (s, "glProgramLocalParameter4fARB");
+	s->getProgramiv = (GLGetProgramivProc)
+	    getProcAddress (s, "glGetProgramivARB");
 
-	if (s->genPrograms	     &&
-	    s->deletePrograms	     &&
-	    s->bindProgram	     &&
-	    s->programString	     &&
-	    s->programEnvParameter4f &&
-	    s->programLocalParameter4f)
+	if (s->genPrograms	       &&
+	    s->deletePrograms	       &&
+	    s->bindProgram	       &&
+	    s->programString	       &&
+	    s->programEnvParameter4f   &&
+	    s->programLocalParameter4f &&
+	    s->getProgramiv)
 	    s->fragmentProgram = 1;
     }
 
@@ -2008,6 +2084,10 @@ addScreen (CompDisplay *display,
 	    s->fbo = 1;
     }
 
+    s->textureCompression = 0;
+    if (strstr (glExtensions, "GL_ARB_texture_compression"))
+	s->textureCompression = 1;
+
     fbConfigs = (*s->getFBConfigs) (dpy,
 				    screenNum,
 				    &nElements);
@@ -3751,21 +3831,7 @@ outputDeviceForPoint (CompScreen *s,
 		      int	 x,
 		      int	 y)
 {
-    int i, x1, y1, x2, y2;
-
-    i = s->nOutputDev;
-    while (i--)
-    {
-	x1 = s->outputDev[i].region.extents.x1;
-	y1 = s->outputDev[i].region.extents.y1;
-	x2 = s->outputDev[i].region.extents.x2;
-	y2 = s->outputDev[i].region.extents.y2;
-
-	if (x1 <= x && x2 > x && y1 <= y && y2 > y)
-	    return i;
-    }
-
-    return s->currentOutputDev;
+    return outputDeviceForGeometry (s, x, y, 1, 1, 0);
 }
 
 void
@@ -3960,6 +4026,27 @@ viewportForGeometry (CompScreen *s,
     }
 }
 
+static int
+rectangleOverlapArea (BOX *rect1,
+		      BOX *rect2)
+{
+    int left, right, top, bottom;
+
+    /* extents of overlapping rectangle */
+    left = MAX (rect1->x1, rect2->x1);
+    right = MIN (rect1->x2, rect2->x2);
+    top = MAX (rect1->y1, rect2->y1);
+    bottom = MIN (rect1->y2, rect2->y2);
+
+    if (left > right || top > bottom)
+    {
+	/* no overlap */
+	return 0;
+    }
+
+    return (right - left) * (bottom - top);
+}
+
 int
 outputDeviceForGeometry (CompScreen *s,
 			 int	    x,
@@ -3968,26 +4055,101 @@ outputDeviceForGeometry (CompScreen *s,
 			 int	    height,
 			 int	    borderWidth)
 {
-    int output = s->currentOutputDev;
-    int x1, y1, x2, y2;
+    int overlapAreas[s->nOutputDev];
+    int i, highest, seen, highestScore;
+    int strategy;
+    BOX geomRect;
 
-    width  += borderWidth * 2;
-    height += borderWidth * 2;
+    if (s->nOutputDev == 1)
+	return 0;
 
-    x1 = s->outputDev[output].region.extents.x1;
-    y1 = s->outputDev[output].region.extents.y1;
-    x2 = s->outputDev[output].region.extents.x2;
-    y2 = s->outputDev[output].region.extents.y2;
+    strategy = s->opt[COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS].value.i;
 
-    if (x1 >= x + width  ||
-	y1 >= y + height ||
-	x2 <= x		 ||
-	y2 <= y)
+    if (strategy == OUTPUT_OVERLAP_MODE_SMART)
     {
-	output = outputDeviceForPoint (s, x + width  / 2, y + height / 2);
+	/* for smart mode, calculate the overlap of the whole rectangle
+	   with the output device rectangle */
+	geomRect.x2 = width + 2 * borderWidth;
+	geomRect.y2 = height + 2 * borderWidth;
+
+	geomRect.x1 = x % s->width;
+	if ((geomRect.x1 + geomRect.x2 / 2) < 0)
+	    geomRect.x1 += s->width;
+	geomRect.y1 = y % s->height;
+	if ((geomRect.y1 + geomRect.y2 / 2) < 0)
+	    geomRect.y1 += s->height;
+
+	geomRect.x2 += geomRect.x1;
+	geomRect.y2 += geomRect.y1;
+    }
+    else
+    {
+	/* for biggest/smallest modes, only use the window center to determine
+	   the correct output device */
+	geomRect.x1 = (x + (width / 2) + borderWidth) % s->width;
+	if (geomRect.x1 < 0)
+	    geomRect.x1 += s->width;
+	geomRect.y1 = (y + (height / 2) + borderWidth) % s->height;
+	if (geomRect.y1 < 0)
+	    geomRect.y1 += s->height;
+
+	geomRect.x2 = geomRect.x1 + 1;
+	geomRect.y2 = geomRect.y1 + 1;
     }
 
-    return output;
+    /* get amount of overlap on all output devices */
+    for (i = 0; i < s->nOutputDev; i++)
+	overlapAreas[i] = rectangleOverlapArea (&s->outputDev[i].region.extents,
+						&geomRect);
+
+    /* find output with largest overlap */
+    for (i = 0, highest = 0, highestScore = 0; i < s->nOutputDev; i++)
+	if (overlapAreas[i] > highestScore)
+	{
+	    highest = i;
+	    highestScore = overlapAreas[i];
+	}
+
+    /* look if the highest score is unique */
+    for (i = 0, seen = 0; i < s->nOutputDev; i++)
+	if (overlapAreas[i] == highestScore)
+	    seen++;
+
+    if (seen > 1)
+    {
+	/* it's not unique, select one output of the matching ones and use the
+	   user preferred strategy for that */
+	unsigned int currentSize, bestOutputSize;
+	Bool         searchLargest;
+	
+	searchLargest = (strategy != OUTPUT_OVERLAP_MODE_PREFER_SMALLER);
+	if (searchLargest)
+	    bestOutputSize = 0;
+	else
+	    bestOutputSize = UINT_MAX;
+
+	for (i = 0, highest = 0; i < s->nOutputDev; i++)
+	    if (overlapAreas[i] == highestScore)
+	    {
+		BOX  *box = &s->outputDev[i].region.extents;
+		Bool bestFit;
+
+		currentSize = (box->x2 - box->x1) * (box->y2 - box->y1);
+
+		if (searchLargest)
+		    bestFit = (currentSize > bestOutputSize);
+		else
+		    bestFit = (currentSize < bestOutputSize);
+
+		if (bestFit)
+		{
+		    highest = i;
+		    bestOutputSize = currentSize;
+		}
+	    }
+    }
+    
+    return highest;
 }
 
 Bool
